from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union, cast

import operator
import sys

from collections import defaultdict
from contextlib import contextmanager
from functools import reduce
from io import StringIO
from pathlib import Path

import numpy as np

from attr import attrib, attrs
from click import progressbar
from mckit.constants import MCNP_ENCODING
from mckit.utils import filter_dict

from .body import Body, Shape
from .box import GLOBAL_BOX, Box
from .card import Card
from .material import Composition, Material
from .surface import Plane, Surface
from .transformation import Transformation
from .utils import accept, on_unknown_acceptor

__all__ = [
    "Universe",
    "produce_universes",
    "NameClashError",
    "cell_selector",
    "surface_selector",
    "collect_transformations",
    "UniverseAnalyser",
]

from .utils.indexes import IndexOfNamed, StatisticsCollector
from .utils.named import Name


class NameClashError(ValueError):
    def __init__(self, result: Union[str, Dict[str, Dict[int, Set["Universe"]]]]) -> None:
        if isinstance(result, str):
            ValueError.__init__(self, result)
        else:
            msg = StringIO()
            msg.write("\n")
            for kind, index in result.items():
                for i, u in index.items():
                    universes = reduce(lambda a, b: a.append(b) or a, map(Universe.name, u), [])
                    msg.write(f"{kind} {i} is found in universes {universes}\n")
            ValueError.__init__(self, msg.getvalue())


def cell_selector(cell_names):
    """Produces cell selector function for specific cell names.

    Parameters
    ----------
    cell_names : int or iterable
        Names of cells to be selected.

    Returns
    -------
    selector : func
        Selector function.
    """
    if isinstance(cell_names, int):
        cell_names = {cell_names}
    else:
        cell_names = set(cell_names)

    def selector(cell):
        if cell.name() in cell_names:
            return [cell]
        else:
            return []

    return selector


def surface_selector(surface_names):
    """Produces surface selector function for specific surface names.

    Parameters
    ----------
    surface_names : int or iterable
        Names of surfaces to be selected.

    Returns
    -------
    selector : func
        Selector function
    """
    if isinstance(surface_names, int):
        surface_names = {surface_names}
    else:
        surface_names = set(surface_names)

    def selector(cell):
        surfs = cell.shape.get_surfaces()
        return [s for s in surfs if s.name() in surface_names]

    return selector


class Universe:
    """Describes universe - a set of cells.

    Universe is a set of cells from which it consist of. Each cell can be filled
    with other universe. In this case, cells of other universe are bounded by
    cell being filled.

    Parameters
    ----------
    cells : iterable
        A list of cells this universe consist of.
    name : int
        Name of the universe, usually used in MCNP. 0 - means global universe.
    verbose_name : str
        Optional verbose name. This name also can be used to retrieve inner
        universe.
    comment : list[str] or str
        String or list of strings that describes the universe.
    common_materials : set
        A set of common materials. Default: None.

    Methods
    -------
    add_cells(cell)
        Adds new cell to the universe.
    apply_fill(cell, universe)
        Applies fill operations to all or selected cells or universes.
    bounding_box(tol, box)
        Gets bounding box of the universe.
    copy()
        Makes a copy of the universe.
    find_common_materials()
        Finds all common materials among universes.
    get_surfaces()
        Gets all surfaces of the universe.
    get_materials()
        Gets all materials of the universe.
    get_compositions()
        Gets all compositions of the universe.
    get_transformations()
        Gets all transformations of the universe.
    get_universes()
        Gets all inner universes.
    name()
        Gets numeric name of the universe.
    name_clashes()
        Checks if there is name clashes.
    rename(start_cell, start_surf, start_mat, start_tr)
        Renames all entities contained in the universe.
    save(filename)
        Saves the universe to file.
    select(cell, selector)
        Selects specified entities.
    set_common_materials(com_mat)
        Sets new common materials for universe and all nested universes.
    simplify(box, split_disjoint, min_volume)
        Simplifies all cells of the universe.
    test_points(points)
        Tests to which cell each point belongs.
    transform(tr)
        Applies transformation tr to this universe. Returns a new universe.
    verbose_name()
        Gets verbose name of the universe.
    """

    def __init__(
        self,
        cells,
        name: Name = 0,
        verbose_name: str = None,
        comment: str = None,
        name_rule: str = "keep",
        common_materials: Set[Composition] = None,
    ):
        self._name = name
        self._comment = comment
        self._verbose_name = verbose_name
        self._cells = []
        if common_materials is None:
            common_materials = set()
        self._common_materials = common_materials

        self.add_cells(cells, name_rule=name_rule)

    @property
    def cells(self):
        return self._cells

    def __iter__(self):
        return iter(self._cells)

    def __len__(self):
        return len(self._cells)

    def __setitem__(self, key: int, value: Body):
        raise NotImplementedError("Renaming rules should be applied.")

    def __getitem__(self, item):
        return self._cells.__getitem__(item)

    def __str__(self):
        return f"Universe(name={self.name()})"

    #
    # dvp: this doesn't work with dictionaries
    #
    # def __hash__(self):
    #     return reduce(xor, map(hash, self.cells), 0)
    #
    # def __eq__(self, other):
    #     return reduce(and_, map(eq, zip(self._cells, other.cells), True))

    def has_equivalent_cells(self, other):
        if len(self) != len(other):
            return False
        for i, c in enumerate(self):
            if not c.is_equivalent_to(other[i]):
                return False
        return True

    def add_cells(self, cells, name_rule="new"):
        """Adds new cell to the universe.

        Modifies current universe.

        Parameters
        ----------
        cells : Body or list[Body]
            An array of cells to be added to the universe.
        name_rule : str
            Rule, what to do with entities' names. 'keep' - keep all names; in
            case of clashes NameClashError exception is raised. 'clash' - rename
            entity only in case of name clashes. 'new' - set new sequential name
            to all inserted entities.
        """
        if isinstance(cells, Body):
            cells = [cells]
        surfs = self.get_surfaces()
        surf_replace = {s: s for s in surfs}
        comps = self._common_materials.union(self.get_compositions())
        comp_replace = {c: c for c in comps}
        cell_names = {c.name() for c in self}
        surf_names = {s.name() for s in surfs}
        comp_names = {c.name() for c in comps}
        for cell in cells:
            if cell.shape.is_empty():
                continue

            new_shape = self._get_cell_replaced_shape(cell, surf_replace, surf_names, name_rule)

            new_cell = Body(new_shape, **cell.options)
            mat = new_cell.material()
            if mat:
                new_comp = Universe._update_replace_dict(
                    mat.composition, comp_replace, comp_names, name_rule, "Material"
                )
                new_cell.options["MAT"] = Material(composition=new_comp, density=mat.density)

            if name_rule == "keep" and cell.name() in cell_names:
                raise NameClashError("Cell name clash: {0}".format(cell.name()))
            elif name_rule == "new" or name_rule == "clash" and cell.name() in cell_names:
                new_name = max(cell_names, default=0) + 1
                new_cell.rename(new_name)
            cell_names.add(new_cell.name())
            new_cell.options["U"] = self
            self._cells.append(new_cell)

    def set_common_materials(self, common_materials):
        """Sets common materials for this one and all nested universes.

        Parameters
        ----------
        common_materials : set
            A set of common materials.
        """
        universes = self.get_universes()
        for u in universes:
            u._set_common_materials(common_materials)

    def _set_common_materials(self, common_materials):
        """Sets common materials for individual universe.

        Parameters
        ----------
        common_materials : set
            A set of common_materials
        """
        self._common_materials = common_materials
        cmd = {m: m for m in common_materials}
        for c in self:
            mat = c.material()
            if mat:
                comp = mat.composition
                if comp in common_materials:
                    c.options["MAT"] = Material(composition=cmd[comp], density=mat.density)

    @staticmethod
    def _get_cell_replaced_shape(cell, surf_replace, surf_names, name_rule):
        cell_surfs = cell.shape.get_surfaces()
        replace_dict = {}
        for s in cell_surfs:
            if isinstance(s, Plane):
                rev_s = Plane(-s._v, -s._k)
                if rev_s in surf_replace.keys():
                    rev_s = surf_replace[rev_s]
                    replace_dict[s] = Shape("C", rev_s)
                    continue
            replace_dict[s] = Universe._update_replace_dict(
                s, surf_replace, surf_names, name_rule, "Surface"
            )
        return cell.shape.replace_surfaces(replace_dict)

    @staticmethod
    def _update_replace_dict(entity, replace, names, rule, err_desc):
        if entity not in replace.keys():
            new_entity = entity.copy()
            if rule == "keep" and new_entity.name() in names:
                print(entity.mcnp_repr())
                for c in replace.keys():
                    print(c.mcnp_repr())
                raise NameClashError("{0} name clash: {1}".format(err_desc, entity.name()))
            elif rule == "new" or rule == "clash" and new_entity.name() in names:
                new_name = max(names, default=0) + 1
                new_entity.rename(new_name)
                names.add(new_name)
            replace[new_entity] = new_entity
            names.add(new_entity.name())
            return new_entity
        else:
            return replace[entity]

    @staticmethod
    def _fill_check(predicate):
        def _predicate(cell):
            fill = cell.options.get("FILL", None)
            if fill:
                return predicate(cell)
            else:
                return False

        return _predicate

    def alone(self):
        """Gets this universe alone, without inner universes.

        Returns
        -------
        u : Universe
            A copy of the universe with FILL cards removed.
        """
        cells = []
        for c in self:
            options = {k: v for k, v in c.options.items() if k != "FILL"}
            cells.append(Body(c.shape, **options))
        return Universe(cells)

    def apply_fill(
        self,
        cell: Union[Body, int] = None,
        universe: Union["Universe", int] = None,
        predicate: Callable[[Body], bool] = None,
        name_rule: str = "new",
    ):
        """Applies fill operations to all or selected cells or universes.

        Modifies current universe.

        Parameters
        ----------
        cell : Body or int
            Cell or name of cell which is filled by filling universe. The cell
            can only belong to this universe. Cells of inner universes are not
            taken into account.
        universe : Universe or int
            Filler-universe or its name. Cells, that have this universe as a
            filler will be filled. Only cells of this universe will be checked.
        predicate : func
            Function that accepts Body instance and return True, if this cell
            must be filled.
        """
        if not cell and not universe and not predicate:

            def predicate(_c):
                return True

        elif cell:

            def predicate(_c):
                return _c.name() == cell

        elif universe:

            def predicate(_c):
                return _c.options["FILL"]["universe"].name() == universe

        predicate = self._fill_check(predicate)
        extra_cells = []
        del_indices = []
        for i, c in enumerate(self):
            if predicate(c):
                extra_cells.extend(c.fill())
                del_indices.append(i)
        for i in reversed(del_indices):
            self._cells.pop(i)
        self.add_cells(extra_cells, name_rule=name_rule)

    def bounding_box(self, tol=100, box=GLOBAL_BOX):
        """Gets bounding box for the universe.

        It finds all bounding boxes for universe cells and then constructs
        box, that covers all cells' bounding boxes. The main purpose of this
        method is to find boundaries of the model geometry to compare
        transformation and surface objects for equality. It is recommended to
        choose tol value not too small to reduce computation time.

        Parameters
        ----------
        tol : float
            Linear tolerance for the bounding box. The distance [in cm] between
            every box's surface and universe in every direction won't exceed
            this value. Default: 100 cm.
        box : Box
            Starting box for the search. The user must be sure that box covers
            all geometry, because cells, that already contains corners of the
            box will be considered as infinite and will be excluded from
            analysis.

        Returns
        -------
        bbox : Box
            Universe bounding box.
        """
        boxes = []
        for c in self._cells:
            test = c.shape.test_points(box.corners)
            if np.any(test == +1):
                continue
            boxes.append(c.shape.bounding_box(tol=tol, box=box))
        all_corners = np.empty((8 * len(boxes), 3))
        for i, b in enumerate(boxes):
            all_corners[i * 8 : (i + 1) * 8, :] = b.corners
        min_pt = np.min(all_corners, axis=0)
        max_pt = np.max(all_corners, axis=0)
        center = 0.5 * (min_pt + max_pt)
        dims = max_pt - min_pt
        return Box(center, *dims)

    def copy(self):
        """Makes a copy of the universe."""
        return Universe(
            self._cells,
            name=self._name,
            verbose_name=self._verbose_name,
            comment=self._comment,
            common_materials=self._common_materials,
        )

    def find_common_materials(self):
        """Finds common materials among universes included.

        Returns
        -------
        common_mats : set
            A set of common materials.
        """
        comp_count = defaultdict(lambda: 0)
        for u in self.get_universes():
            for c in u.get_compositions():
                comp_count[c] += 1
        common_mats = {c for c, cnt in comp_count.items() if cnt > 1}
        return common_mats

    def get_surfaces(self, inner: bool = False) -> Set[Surface]:
        """Gets all surfaces of the universe.

        Parameters
        ----------
        inner : bool
            Whether to take surfaces of inner universes. Default: False -
            return surfaces of this universe only.

        Returns
        -------
        surfs : set
            A set of surfaces that belong to the universe.
        """
        surfs = set()
        for c in self:
            surfs.update(c.shape.get_surfaces())
            if inner and "FILL" in c.options.keys():
                surfs.update(c.options["FILL"]["universe"].get_surfaces(inner))
        return surfs

    def get_surfaces_list(self, inner: bool = False):
        def reducer(surfaces_list, cell):
            surfaces_list.extend(cell.shape.get_surfaces())
            if inner and "FILL" in cell.options:
                surfaces_list.extend(cell.options["FILL"]["universe"].get_surfaces_list(inner))
            return surfaces_list

        return reduce(reducer, self, [])

    def get_compositions(self, exclude_common: bool = False) -> Set[Composition]:
        """Gets all compositions of the universe.

        Parameters
        ----------
        exclude_common : bool
            Exclude common compositions from the result. Default: False.

        Returns
        -------
        comps : set
            A set of Composition objects.
        """
        compositions = set()
        for c in self:
            material = c.material()
            if material:
                compositions.add(material.composition)
        if exclude_common:
            compositions.difference_update(self._common_materials)
        return compositions

    def get_universes(self) -> Set["Universe"]:
        """Gets all inner universes.

        Returns
        -------
        universes : set
            A set of universes.
        """
        universes = {self}
        for c in self:
            if "FILL" in c.options:
                u = c.options["FILL"]["universe"]
                universes.update(u.get_universes())
        return universes

    def name(self) -> Name:
        """Gets numeric name of the universe."""
        return self._name

    def name_clashes(self) -> Dict[str, Dict[int, Set["Universe"]]]:
        """Checks, if there is name clashes.

        Returns
        -------
        stat : dict
            Description of found clashes. If no clashes - the dictionary is empty.
        """
        universes = self.get_universes()
        universe_to_cell_name_map = {u: list(map(Card.name, u)) for u in universes}
        universe_to_surface_name_map = {
            u: list(map(Card.name, u.get_surfaces())) for u in universes
        }
        mats = {None: list(map(Card.name, self._common_materials))}
        for u in universes:
            mats[u] = list(map(Card.name, u.get_compositions().difference(self._common_materials)))
        univs = {u: [u.name()] for u in universes}
        cstat = Universe._produce_stat(universe_to_cell_name_map)
        stat = {}
        if cstat:
            stat["cell"] = cstat
        sstat = Universe._produce_stat(universe_to_surface_name_map)
        if sstat:
            stat["surf"] = sstat
        mstat = Universe._produce_stat(mats)
        if mstat:
            stat["material"] = mstat
        ustat = Universe._produce_stat(univs)
        if ustat:
            stat["universe"] = ustat
        # TODO dvp: handle transformations here
        return stat

    @staticmethod
    def _produce_stat(names: Dict["Universe", Iterable[int]]) -> Dict[int, Set["Universe"]]:
        stat = defaultdict(list)
        for u, u_names in names.items():
            for name in u_names:
                stat[name].append(u)
        return Universe._clean_stat_dict(stat)

    @staticmethod
    def _clean_stat_dict(stat) -> Dict[int, Set["Universe"]]:
        new_stat = {}
        for k, v in stat.items():
            if len(v) > 1:
                new_stat[k] = set(v)
        return new_stat

    def rename(
        self,
        start_cell: int = None,
        start_surf: int = None,
        start_mat: int = None,
        start_tr: int = None,
        name: int = None,
    ) -> None:
        """Renames all entities contained in the universe.

        All new names are sequential starting from the specified name. If name
        is None, than names of entities are leaved untouched.

        Parameters
        ----------
        start_cell :
            Starting name for cells. Default: None.
        start_surf :
            Starting name for surfaces. Default: None.
        start_mat :
            Starting name for materials. Default: None.
        start_tr :
            Starting name for transformations. Default: None.
        name :
            Name for the universe. Default: None.
        """
        # TODO dvp: implement transformations renaming
        assert start_tr is None, "Transformation renaming is not implemented yet"
        if name:
            self._name = name
            for c in self:
                c.options = filter_dict(c.options, "original")
        if start_cell:
            for c in self:
                c.rename(start_cell)
                start_cell += 1
        if start_surf:
            surfs = self.get_surfaces()
            for s in sorted(surfs, key=Card.name):
                s.rename(start_surf)
                start_surf += 1
        if start_mat:
            mats = self.get_compositions()
            for m in sorted(mats, key=Card.name):
                if m not in self._common_materials:
                    m.rename(start_mat)
                    start_mat += 1

    def check_clashes(self) -> None:
        result = self.name_clashes()
        if result:
            raise NameClashError(result)

    def save(
        self,
        filename: Union[str, Path],
        encoding: str = MCNP_ENCODING,
        check_clashes: bool = True,
    ):
        """Saves the universe into file."""
        # NOTE dvp: Don't try to resolve names here, the object shouldn't change on save() function.
        # self.check_clashes()
        if check_clashes:
            analyser = UniverseAnalyser(self)
            if not analyser.we_are_all_clear():
                out = StringIO()
                out.write("Duplicates found:\n")
                analyser.print_duplicates_map(stream=out)
                raise NameClashError(out.getvalue())

        transformations = collect_transformations(self)
        if transformations:
            transformations = sorted(transformations, key=Card.name)
        universes = self.get_universes()
        cells = []
        surfaces = []
        materials = list(sorted(self._common_materials, key=Card.name))
        for u in sorted(universes, key=Universe.name):
            cells.extend(sorted(u, key=Card.name))
            surfaces.extend(sorted(u.get_surfaces(), key=Card.name))
            materials.extend(sorted(u.get_compositions(True), key=Card.name))
        cards = [self.verbose_name]
        cards.extend(map(Card.mcnp_repr, cells))
        cards.append("")
        cards.extend(map(Card.mcnp_repr, surfaces))
        cards.append("")
        if transformations:
            cards.extend(map(Card.mcnp_repr, transformations))
        if materials:
            cards.extend(map(Card.mcnp_repr, materials))
        cards.append("")
        with open(filename, mode="w", encoding=encoding) as f:
            f.write("\n".join(cards))

    def select(self, selector=None, inner=False):
        """Selects specified entities.

        Parameters
        ----------
        selector : func
            A function that accepts 1 argument, Body instance, and returns
            selected entities.
        inner : bool
            Whether to consider inner universes. Default: False - only this
            universe will be taken into account.

        Returns
        -------
        items : list
            List of selected items.
        """
        items = []
        taken_ids = set()
        for c in self:
            portion = selector(c)
            if inner:
                u = c.options.get("FILL", {}).get("universe", None)
                if u:
                    portion.extend(u.select(selector, True))
            for item in portion:
                if id(item) not in taken_ids:
                    taken_ids.add(id(item))
                    items.append(item)
        return items

    def simplify(self, box=GLOBAL_BOX, min_volume=1, split_disjoint=False, verbose=True):
        """Simplifies all cells of the universe.

        Modifies current universe.

        Parameters
        ----------
        box : Box
            Box, from which simplification process starts. Default: GLOBAL_BOX.
        min_volume : float
            Minimal volume of the box, when splitting process terminates.
        split_disjoint : bool
            Whether to split disjoint cells.
        verbose : bool
            Turns on verbose output. Default: True.
        """
        new_cells = []
        if verbose:

            def fmt_fun(x):
                return "Simplifying cell #{0}".format(x.name() if x else x)

            uiter = progressbar(self, item_show_func=fmt_fun).__enter__()
        else:
            uiter = self

        for c in uiter:
            cs = c.simplify(box=box, min_volume=min_volume)
            if not cs.shape.is_empty():
                new_cells.append(cs)

        if verbose:
            print("Universe {0} simplification has been finished.".format(self.name()))
            print("{0} empty cells were deleted.".format(len(self._cells) - len(new_cells)))

        self._cells = new_cells

    def test_points(self, points):
        """Finds cell to which each point belongs to.

        Parameters
        ----------
        points : array_like[float]
            An array of point coordinates. If there is only one point it has
            shape (3,); if there are n points, it has shape (n, 3).

        Returns
        -------
        result : np.ndarray[int]
            An array of cell indices to which a particular point belongs to.
            Its length equals to the number of points.
        """
        points = np.array(points)
        result = np.empty(points.size // 3)
        for i, c in enumerate(self._cells):
            test = c.shape.test_points(points)
            result[test == +1] = i
        return result

    def transform(self, tr: Transformation) -> "Universe":
        """Applies transformation tr to this universe.

        Returns a new universe.
        """
        new_cells = [c.transform(tr) for c in self]
        return Universe(
            new_cells,
            name=self._name,
            name_rule="clash",
            verbose_name=self._verbose_name,
            comment=self._comment,
        )

    def apply_transformation(self) -> "Universe":
        """Applies transformations specified in cells.

        Returns a new universe.
        """
        new_cells = [c.apply_transformation() for c in self]
        return Universe(
            new_cells,
            name=self._name,
            name_rule="clash",
            verbose_name=self._verbose_name,
            comment=self._comment,
        )

    @property
    def verbose_name(self) -> str:
        """Gets verbose name of the universe."""
        return str(self.name()) if self._verbose_name is None else self._verbose_name

    @property
    def comment(self):
        return self._comment


@attrs
class _UniverseCellsGroup:
    universe: Universe = attrib()
    cells: List[Body] = attrib()


def produce_universes(cells: Iterable[Body]) -> Universe:
    """Creates groups from cells.

    The function groups all the cells by 'universe' option value,
    and creates corresponding groups.

    Parameters
    ----------
    cells : Iterable[Body]
        Cells to process.

    Returns
    -------
    universe : Universe
        The top level universe with name = 0.
    """
    groups: Dict[Name, _UniverseCellsGroup] = {}
    for c in cells:
        universe_no: Name = c.options.get("U", 0)
        if universe_no in groups:
            groups[universe_no].cells.append(c)
        else:
            new_group = _UniverseCellsGroup(universe=Universe([], universe_no), cells=[c])
            groups[universe_no] = new_group
    for c in cells:
        fill: Dict[str, Any] = c.options.get("FILL", None)
        if fill is not None:
            fill_universe_no = fill["universe"]
            fill["universe"] = groups[fill_universe_no].universe
    for group in groups.values():
        group.universe.add_cells(group.cells, name_rule="keep")
    top_universe = groups[0].universe
    top_universe.set_common_materials(top_universe.find_common_materials())
    return top_universe


def collect_transformations(universe: Universe, recursive=True) -> Set[Transformation]:
    def add_surface_transformation(aggregator: Set[Transformation], surface: Surface) -> None:
        transformation = surface.transformation
        if transformation and transformation.name():
            aggregator.add(transformation)

    def at_surface(aggregator: Set[Transformation], s: Union[Surface, Shape, Body]):
        if isinstance(s, Surface):
            add_surface_transformation(aggregator, s)
        elif isinstance(s, Shape):
            aggregator.update(accept(s, visit_shape))
        else:
            assert isinstance(s, Body)
            aggregator.update(accept(s, visit_body))
        return aggregator

    @contextmanager
    def visit_shape(s: Union[Surface, Body, Shape]):
        if isinstance(s, (Surface, Body, Shape)):
            yield at_surface, set()
        else:
            on_unknown_acceptor(s)

    def at_shape(aggregator: Set[Transformation], s: Union[Surface, Body, Shape]):
        if isinstance(s, Surface):
            add_surface_transformation(aggregator, s)
            return aggregator
        aggregator.update(accept(s, visit_shape))
        return aggregator

    @contextmanager
    def visit_body(b: Body):
        if isinstance(b, Body):
            yield at_shape, set()
        else:
            on_unknown_acceptor(b)

    def at_body(aggregator: Set[Transformation], b: Body) -> Set[Transformation]:
        body_transformation = b.transformation
        if body_transformation and body_transformation.name():
            aggregator.add(body_transformation)
        if recursive:
            fill = b.options.get("FILL", None)
            if fill:
                fill_universe = fill["universe"]
                fill_transformation = fill.get("transform", None)
                if fill_transformation and fill_transformation.name():
                    aggregator.add(fill_transformation)
                aggregator.update(collect_transformations(fill_universe))
        aggregator.update(accept(b, visit_body))
        return aggregator

    @contextmanager
    def visit_universe(u: Universe) -> Set[Transformation]:
        if isinstance(u, Universe):
            yield at_body, set()
            # TODO dvp:  set() is not a valid choice as aggregator considering
            #        cases when names differ but transformations are equal by values
        else:
            on_unknown_acceptor(u)

    return accept(universe, visit_universe)


# TODO dvp: it's possible to generalize visiting introducing class Visitor
#           See: all visit_... functions will be probably not changed on deriving
#           Use functools partial to hide self parameter when passing the method as a function to accept()


# TODO dvp: make names of cards not optional
IU = Tuple[List[Optional[Name]], Name]
"""Entities, Universe name."""

E2U = Dict[Name, Dict[Name, int]]
"""Map Entity name -> Universe Name -> Count."""


def entity_to_universe_map_reducer(result: E2U, entry: IU) -> E2U:
    entities_names, universe_name = entry

    def inner_reducer(_result: E2U, entity_name: Name) -> E2U:
        _result[entity_name][universe_name] += 1
        return _result

    return reduce(inner_reducer, entities_names, result)


def cells_to_universe_mapper(universe: Universe) -> IU:
    return list(map(Body.name, universe)), universe.name()


def surfaces_to_universe_mapper(universe: Universe) -> IU:
    return (
        list(map(Surface.name, universe.get_surfaces_list(inner=False))),
        universe.name(),
    )


def compositions_to_universe_mapper(universe: Universe) -> IU:
    return (
        list(map(Composition.name, universe.get_compositions(exclude_common=False))),
        universe.name(),
    )


def transformations_to_universe_mapper(universe: Universe) -> IU:
    return (
        list(map(Transformation.name, collect_transformations(universe, recursive=False))),
        universe.name(),
    )


def is_shared_between_universes(item: Tuple[Name, Dict[Name, int]]) -> bool:
    entity, universes_counts = item
    return 1 < len(universes_counts.keys())


def make_universe_counter_map():
    return defaultdict(int)


def collect_shared_entities(
    entities_extractor: Callable[[Universe], IU], universes: Iterable[Universe]
) -> E2U:
    return dict(
        filter(
            is_shared_between_universes,
            reduce(
                entity_to_universe_map_reducer,
                map(entities_extractor, universes),
                cast(E2U, defaultdict(make_universe_counter_map)),
            ).items(),
        )
    )


class UniverseAnalyser:
    def __init__(self, universe: Universe):
        self.universe = universe
        universes = self.universe.get_universes()
        self.universe_duplicates = StatisticsCollector(ignore_equal=True)
        self.universes_index = IndexOfNamed.from_iterable(
            universes,
            on_duplicate=self.universe_duplicates,
        )
        self.cell_duplicates = StatisticsCollector()
        cells: List[Body] = reduce(operator.add, map(list, universes), [])
        self.cell_index = IndexOfNamed[Name, Body].from_iterable(
            cells,
            on_duplicate=self.cell_duplicates,
        )
        self.cell_to_universe_map = collect_shared_entities(cells_to_universe_mapper, universes)
        self.surface_duplicates = StatisticsCollector(ignore_equal=True)
        self.surface_index = IndexOfNamed[Name, Surface].from_iterable(
            universe.get_surfaces_list(inner=True),
            on_duplicate=self.surface_duplicates,
        )
        self.surface_to_universe_map = collect_shared_entities(
            surfaces_to_universe_mapper, universes
        )
        self.composition_duplicates = StatisticsCollector(ignore_equal=True)
        self.composition_index = IndexOfNamed[Name, Composition].from_iterable(
            universe.get_compositions(),
            on_duplicate=self.composition_duplicates,
        )
        self.compositions_to_universe_map = collect_shared_entities(
            compositions_to_universe_mapper, universes
        )
        self.transformation_duplicates = StatisticsCollector(ignore_equal=True)
        self.transformation_index = IndexOfNamed[Name, Transformation].from_iterable(
            collect_transformations(universe, recursive=True),
            on_duplicate=self.transformation_duplicates,
        )
        self.transformations_to_universe_map = collect_shared_entities(
            transformations_to_universe_mapper, universes
        )

    def duplicates(self):
        return (
            self.cell_duplicates,
            self.surface_duplicates,
            self.composition_duplicates,
            self.transformation_duplicates,
        )

    def duplicates_maps(self):
        return (
            self.cell_to_universe_map,
            self.surface_to_universe_map,
            self.compositions_to_universe_map,
            self.transformations_to_universe_map,
        )

    def we_are_all_clear(self) -> bool:
        return not any(self.duplicates())

    def print_duplicates_map(self, stream=sys.stdout):
        def printer(_, item: Tuple[str, E2U]):
            tag, info = item
            entities = sorted(list(info.keys()))
            for e in entities:
                universes_count_map = info[e]
                universes = sorted(list(universes_count_map.keys()))
                print(f"{tag} {e} occurs", file=stream)
                for u in universes:
                    print(
                        f"   in universe {u} {universes_count_map[u]} times",
                        file=stream,
                    )

        reduce(
            printer,
            zip(
                ["cell", "surface", "composition", "transformation"],
                self.duplicates_maps(),
            ),
            None,
        )



from typing import Optional

import mckit.parser.common.utils as pu
import sly

from mckit.body import Body, Shape
from mckit.material import Material
from mckit.parser.common import CellStrictIndex, CompositionStrictIndex
from mckit.parser.common import Lexer as LexerBase
from mckit.parser.common import SurfaceStrictIndex, TransformationStrictIndex
from mckit.surface import Surface
from mckit.transformation import Transformation
from mckit.utils import filter_dict
from mckit.utils.indexes import Index

CELL_WORDS = {"U", "MAT", "LAT", "TMP", "RHO", "VOL", "PMT"}


def intern_cell_word(word: str):
    word = pu.ensure_upper(word)
    word, res = pu.internalize(word, CELL_WORDS)
    if not res:
        raise pu.ParseError(f"'{word}' is not a valid word for cell specification")
    return word


# noinspection PyPep8Naming,PyUnboundLocalVariable,PyUnresolvedReferences,SpellCheckingInspection
class Lexer(LexerBase):
    literals = {":", "(", ")", "*", "#"}
    ignore = "[ \t,=]"
    tokens = {
        INT_ATTR,
        IMP,
        FLOAT_ATTR,
        TRCL,
        FILL,
        INTEGER,
        FLOAT,
        ZERO,
        LIKE,
        BUT,
        N,
        P,
        E,
    }
    INT_ATTR = "U|MAT|LAT|PMT"
    IMP = "IMP"
    FLOAT_ATTR = "TMP|RHO|VOL"
    TRCL = "TRCL"
    FILL = "FILL"
    LIKE = "LIKE"
    BUT = "BUT"
    N = "N"
    P = "P"
    E = "E"

    @_(pu.FLOAT)
    def FLOAT(self, token):
        return self.on_float(token)


# noinspection PyUnresolvedReferences
class Parser(sly.Parser):
    tokens = Lexer.tokens

    def __init__(
        self,
        cells,
        surfaces,
        transformations,
        compositions,
        comments=None,
        trailing_comments=None,
        original=None,
    ):
        sly.Parser.__init__(self)
        self._cells = cells
        self._surfaces = surfaces
        self._transformations = transformations
        self._compositions = compositions
        self._comments = comments
        self._trailing_comments = trailing_comments
        self._original = original

    @property
    def cells(self):
        return self._cells

    @property
    def surfaces(self):
        return self._surfaces

    @property
    def transformations(self):
        return self._transformations

    @property
    def compositions(self):
        return self._compositions

    @property
    def comments(self):
        return self._comments

    @property
    def trailing_comments(self):
        return self._trailing_comments

    @property
    def original(self):
        return self._original

    def build_cell(self, geometry, options):
        if self.trailing_comments:
            options["comment"] = self.trailing_comments
        if self._original:
            options["original"] = self.original
        return Body(geometry, **options)

    @_("INTEGER cell_material cell_spec")
    def cell(self, p):
        geometry, options = p.cell_spec  # type: TGeometry, Optional[Dict[str, Any]]
        assert options is None or isinstance(options, dict)
        if options is None:
            options = {"name": p.INTEGER}
        else:
            options["name"] = p.INTEGER
        if p.cell_material is not None:
            composition_no, density = p.cell_material  # type: int, float
            composition = self.compositions[composition_no]
            if density > 0:
                material = Material(composition=composition, concentration=density * 1e24)
            else:
                material = Material(composition=composition, density=-density)
            options["MAT"] = material
        return self.build_cell(geometry, options)

    @_("INTEGER LIKE INTEGER BUT attributes")
    def cell(self, p):
        reference_body = self.cells[p[2]]
        options = filter_dict(reference_body.options, "original", "comment", "comment_above")
        options.update(p.attributes)
        options["name"] = p[0]
        new_body = Body(reference_body, **options)
        return self.build_cell(new_body, options)

    @_("INTEGER float")
    def cell_material(self, p):
        return p.INTEGER, p.float

    @_("ZERO")
    def cell_material(self, _):
        return None

    @_("expression attributes")
    def cell_spec(self, p):
        return p.expression, p.attributes

    @_("expression")
    def cell_spec(self, p):
        return p.expression, None

    @_('expression ":" term')
    def expression(self, p):
        result = p.expression + p.term
        result.append("U")
        return result

    @_("term")
    def expression(self, p):
        return p.term

    @_("term factor")
    def term(self, p):
        res = p.term + p.factor
        res.append("I")
        return res

    @_("factor")
    def term(self, p):
        return p.factor

    @_('"#" "(" expression ")"')
    def factor(self, p):
        return p.expression + ["C"]

    @_('"(" expression ")"')
    def factor(self, p):
        return p.expression

    @_('"-" integer')
    def factor(self, p):
        surface = self.surfaces[p.integer]
        return [Shape("C", surface)]

    @_('"+" integer')
    def factor(self, p):
        surface = self.surfaces[p.integer]
        return [Shape("S", surface)]

    @_('"#" integer')
    def factor(self, p):
        body = self.cells[p.integer]
        return [Shape("C", body)]

    @_("integer")
    def factor(self, p):
        item: int = p.integer
        if 0 < item:
            opc = "S"
        else:
            opc = "C"
            item = -item
        surface: Surface = self.surfaces[item]
        return [Shape(opc, surface)]

    @_("attributes attribute")
    def attributes(self, p):
        result = p.attributes
        result.update(p.attribute)
        return result

    @_("attribute")
    def attributes(self, p):
        result = p[0]
        return result

    @_(
        "fill_attribute",
        "trcl_attribute",
        "imp_attribute",
        "float_attribute",
        "int_attribute",
    )
    def attribute(self, p):
        return p[0]

    @_('"*" FILL integer "(" transform_params ")"')
    def fill_attribute(self, p):
        translation, rotation, inverted = p.transform_params
        transformation = Parser.build_transformation(translation, rotation, True, inverted)
        fill = {"universe": p.integer, "transform": transformation}
        return {"FILL": fill}

    @_('FILL integer "(" transform_params ")"')
    def fill_attribute(self, p):
        translation, rotation, inverted = p.transform_params
        transformation = Parser.build_transformation(translation, rotation, False, inverted)
        fill = {"universe": p.integer, "transform": transformation}
        return {"FILL": fill}

    @_('FILL integer "(" integer ")"')
    def fill_attribute(self, p):
        transform_id: int = p[3]
        transformation = self.transformations[transform_id]
        fill = {"universe": p[1], "transform": transformation}
        return {"FILL": fill}

    @_("FILL integer")
    def fill_attribute(self, p):
        fill = {"universe": p.integer}
        return {"FILL": fill}

    @staticmethod
    def build_transformation(translation, rotation, in_degrees, inverted):
        return Transformation(
            translation=translation,
            rotation=rotation,
            indegrees=in_degrees,
            inverted=inverted,
        )

    @_('"*" TRCL "(" transform_params ")"')
    def trcl_attribute(self, p):
        translation, rotation, inverted = p.transform_params
        return {"TRCL": Parser.build_transformation(translation, rotation, True, inverted)}

    @_('TRCL "(" transform_params ")"')
    def trcl_attribute(self, p):
        translation, rotation, inverted = p.transform_params
        return {"TRCL": Parser.build_transformation(translation, rotation, False, inverted)}

    @_("TRCL integer")
    def trcl_attribute(self, p):
        transform_id: int = p.integer
        transformation = self.transformations[transform_id]
        return {"TRCL": transformation}

    @_("translation rotation")
    def transform_params(self, p):
        rotation, inverted = p.rotation
        return p.translation, rotation, inverted

    @_("translation")
    def transform_params(self, p):
        return p.translation, None, False

    @_("float float float")
    def translation(self, p):
        return [f for f in p]

    @_("float float float float float float float float float INTEGER")
    def rotation(self, p):
        m = p[9]
        assert m == -1 or m == 1, f"Invalid value for transformation M parameter {m}"
        return [f for f in p][:-1], m == -1

    @_(
        "float float float float float float float float float",
        "float float float float float float",
        "float float float float float",
        "float float float",
    )
    def rotation(self, p):
        return [f for f in p], False

    #
    #  See MCNP, vol.I, p.3-7
    #  Example: IMP:E,P,N 1 1 0.
    #
    @_('IMP ":" particle_list float_list')
    def imp_attribute(self, p):
        number_of_particles = len(p.particle_list)
        if number_of_particles != len(p.float_list):
            while len(p.float_list) < number_of_particles:
                p.float_list.append(p.float_list[-1])
        result = dict(("IMP" + k, v) for k, v in zip(p.particle_list, p.float_list))
        return result

    @_("float_list float")
    def float_list(self, p):
        return p.float_list + [p.float]

    @_("float")
    def float_list(self, p):
        return [p.float]

    @_("particle_list particle")
    def particle_list(self, p):
        p.particle_list.append(p.particle)
        return p.particle_list

    @_("particle")
    def particle_list(self, p):
        return [p.particle]

    @_("N", "P", "E")
    def particle(self, p):
        result = pu.ensure_upper(p[0])
        return result

    @_("FLOAT_ATTR float")
    def float_attribute(self, p):
        return {intern_cell_word(p.FLOAT_ATTR): p.float}

    @_("INT_ATTR integer")
    def int_attribute(self, p):
        return {intern_cell_word(p.INT_ATTR): p.integer}

    @_("FLOAT")
    def float(self, p):
        return p.FLOAT

    @_("integer")
    def float(self, p):
        return float(p[0])

    @_("INTEGER", "ZERO")
    def integer(self, p):
        return p[0]


def parse(
    text: str,
    cells: Optional[Index] = None,
    surfaces: Optional[Index] = None,
    transformations: Optional[Index] = None,
    compositions: Optional[Index] = None,
) -> Body:
    cells, surfaces, transformations, compositions = map(
        lambda x: x[1]() if x[0] is None else x[0],
        zip(
            [cells, surfaces, transformations, compositions],
            [
                CellStrictIndex,
                SurfaceStrictIndex,
                TransformationStrictIndex,
                CompositionStrictIndex,
            ],
        ),
    )
    original = text
    text = pu.drop_c_comments(text)
    text, comments, trailing_comments = pu.extract_comments(text)
    lexer = Lexer()
    parser = Parser(
        cells,
        surfaces,
        transformations,
        compositions,
        comments,
        trailing_comments,
        original,
    )
    result = parser.parse(lexer.tokenize(text))
    return result



from typing import Iterable, Optional

from mckit.surface import EX, Plane, Surface
from mckit.utils.indexes import Index, NumberedItemNotFoundError


class DummySurface(Plane):
    """Dummy surface to substitute surface when it's not found in a ``IgnoringIndex``."""

    def __init__(self, name: int):
        super().__init__(EX, 0.0, name=name, comment="dummy")

    def __str__(self) -> str:
        return str(self.name())

    def __repr__(self):
        return f"DummySurface({self.name()})"


def raise_on_absent_surface_strategy(name: int) -> Optional[DummySurface]:
    raise SurfaceNotFoundError(name)


def dummy_on_absent_surface_strategy(name: int) -> Optional[DummySurface]:
    return DummySurface(name)


class SurfaceStrictIndex(Index):
    def __init__(self, **kwargs):
        super().__init__(raise_on_absent_surface_strategy, **kwargs)

    @classmethod
    def from_iterable(cls, items: Iterable[Surface]) -> "SurfaceStrictIndex":
        index = cls()
        index.update((c.name(), c) for c in items)
        return index


class SurfaceDummyIndex(Index):
    def __init__(self, **kwargs):
        super().__init__(dummy_on_absent_surface_strategy, **kwargs)


class SurfaceNotFoundError(NumberedItemNotFoundError):
    kind = "Surface"



from typing import Iterable, Optional

from mckit.transformation import Transformation
from mckit.utils.indexes import Index, NumberedItemNotFoundError


class DummyTransformation(Transformation):
    """To substitute transformation when it's not found."""

    def __init__(self, name: int):
        super().__init__(name=name, comment="dummy")


def raise_on_absent_transformation_strategy(name: int) -> Optional[DummyTransformation]:
    raise TransformationNotFoundError(name)


def dummy_on_absent_transformation_strategy(name: int) -> Optional[DummyTransformation]:
    return DummyTransformation(name)


class TransformationStrictIndex(Index):
    def __init__(self, **kwargs):
        super().__init__(raise_on_absent_transformation_strategy, **kwargs)

    @classmethod
    def from_iterable(cls, items: Iterable[Transformation]) -> "TransformationStrictIndex":
        index = cls()
        index.update((c.name(), c) for c in items)
        return index


class TransformationDummyIndex(Index):
    def __init__(self, **kwargs):
        super().__init__(dummy_on_absent_transformation_strategy, **kwargs)


class TransformationNotFoundError(NumberedItemNotFoundError):
    kind = "Transformation"



import re

import sly

# noinspection PyProtectedMember
from sly.lex import LexError

IS_INTEGER_REGEX = re.compile(r"[-+]?\d+$")


# noinspection PyUnresolvedReferences,SpellCheckingInspection,PyTypeChecker
class Lexer(sly.Lexer):
    tokens = set()  # Should be defined in all subclasses of sly.Lexer
    literals = {":", "(", ")"}  # ---
    ignore = " \t&"  # most common for MCNP text parsers
    reflags = re.IGNORECASE | re.MULTILINE  # ---

    @staticmethod
    def on_float(token, use_zero=True, use_integer=True):
        if use_integer and IS_INTEGER_REGEX.match(token.value) is not None:
            token = Lexer.on_integer(token, use_zero)
        else:
            token.value = float(token.value)
        return token

    @staticmethod
    def on_integer(token, use_zero=True):
        token.value = int(token.value)
        if use_zero and token.value == 0:
            token.type = "ZERO"
        else:
            token.type = "INTEGER"
        return token

    @_(r"\n+")
    def ignore_newline(self, token):
        self.lineno += len(token.value)

    def get_start_of_line(self, token):
        prev_cr = self.text.rfind("\n", 0, token.index)
        return prev_cr

    def get_end_of_line(self, token):
        next_cr = self.text.find("\n", token.index)
        if next_cr < 0:
            next_cr = len(self.text)
        return next_cr

    @staticmethod
    def column(token, prev_cr):
        return token.index - prev_cr

    def find_column(self, token):
        prev_cr = self.get_start_of_line(token)
        column = Lexer.column(token, prev_cr)
        return column

    def error(self, token):
        prev_cr = self.get_start_of_line(token)
        next_cr = self.get_end_of_line(token)
        column = Lexer.column(token, prev_cr)
        msg = (
            f"Illegal character '{token.value[0]}', at line {self.lineno}, column {self.find_column(token)}\n"
            + f"{self.text[prev_cr + 1: next_cr]}\n"
            + " " * (column - 1)
            + "^"
        )
        raise LexError(msg, token.value, token.index)



import numpy as np

from mckit.utils import get_decades, significant_digits

# def run_digits_in_fraction(a: np.ndarray) -> None:
#     map(digits_in_fraction_for_str, a)
#
#
# def test_digits_in_fraction(benchmark):
#     values = (np.random.rand(1000) - 0.5) * 1000.0
#     benchmark(run_digits_in_fraction, values)

# Name (time in ns)                Min       Max      Mean   StdDev    Median     IQR    Outliers  OPS (Mops/s)  Rounds  Iterations
# test_digits_in_fraction     186.9630  720.5936  194.0950  26.6555  190.4080  1.2957  3901;10481        5.1521  192753          27


def run_significant_digits(a: np.ndarray) -> None:
    map(significant_digits, a)


def test_significant_digits(benchmark):
    values = (np.random.rand(1000) - 0.5) * 1000.0
    benchmark(run_significant_digits, values)


# Name (time in ns)                Min       Max      Mean   StdDev    Median      IQR   Outliers  OPS (Mops/s)  Rounds  Iterations
# test_significant_digits     185.8118  811.3717  200.8212  30.4897  190.1493  21.4075  2785;2766        4.9796  193839          27


def run_get_decades(a: np.ndarray) -> None:
    return np.fromiter(map(get_decades, a), np.int16)


def test_get_decades(benchmark):
    values = (np.random.rand(1000) - 0.5) * 1000.0
    benchmark(run_get_decades, values)


# Name (time in ms)        Min     Max    Mean  StdDev  Median     IQR  Outliers       OPS  Rounds  Iterations
# test_get_decades      2.8199  5.4146  2.9415  0.2459  2.8766  0.0728     12;45  339.9577     331           1



from io import StringIO

from mckit.parser import from_file
from mckit.universe import UniverseAnalyser
from mckit.utils import path_resolver

cli_data = path_resolver("tests.cli")


def test_test_universe_analyser():
    universe_path = cli_data("data/shared_surface.mcnp")
    assert universe_path.exists()
    universe = from_file(universe_path).universe
    analyzer = UniverseAnalyser(universe)
    assert not analyzer.cell_duplicates
    assert not analyzer.surface_duplicates
    assert analyzer.surface_to_universe_map
    assert analyzer.surface_to_universe_map[1] == {
        0: 3,
        1: 1,
    }, "The surface 1 occurs 3 times in universes 0 and once in universe 1"
    assert analyzer.we_are_all_clear()
    out = StringIO()
    analyzer.print_duplicates_map(stream=out)
    assert "surface 1 occurs" in out.getvalue()



# -*- coding: utf-8 -*-

"""Разложение модели на составляющие юниверсы.

Читает модель, извлекает юниверсы первого уровня и сохраняет их
в каталог universes под именем uN.i, где N - номер юниверса.
Также сохраняет общую модель (без юниверсов) под именем envelopes.i
Создает спецификацию для последующей сборки модели в виде TOML файла.
"""
from datetime import datetime
from pathlib import Path

import tomli_w as tw

from mckit import Universe
from mckit.cli.logging import logger
from mckit.parser.mcnp_input_sly_parser import ParseResult, from_file
from mckit.universe import collect_transformations

from .common import save_mcnp


def get_default_output_directory(source):
    return Path(Path(source).with_suffix(".universes").name)


def move_universe_attribute_to_comments(universe):
    no = universe.name()
    for cell in universe:
        cell.options.pop("U", None)
        comm = cell.options.get("comment", [])
        comm.append(f"U={no}")
        cell.options["comment"] = comm


def decompose(output, fill_descriptor_path, source, override):
    logger.info("Running mckit decompose")
    logger.debug("Working dir {}", Path(".").absolute())
    logger.info(f"Processing {source}")
    logger.debug("Loading model from {}", source)
    source = Path(source)
    if output is None:
        output = get_default_output_directory(source)
    else:
        output = Path(output)
    output.mkdir(parents=True, exist_ok=True)
    parse_result: ParseResult = from_file(source)
    model: Universe = parse_result.universe
    named_transformations = list(collect_transformations(model))
    already_processed_universes = set()
    # TODO dvp: check ordering of items in resulting file
    fill_descriptor = {
        "title": parse_result.title,
        "source": source.name,
        "comment": model.comment if model.comment else "",
        "created": datetime.now(),
    }
    for c in model:
        fill = c.options.pop("FILL", None)
        if fill:
            universe = fill["universe"]
            words = [f"FILL={universe.name()}"]
            transform = fill.get("transform", None)
            if transform:
                words[0] = "*" + words[0]
                words.append("(")
                words.extend(transform.get_words())
                words.append(")")
            comm = c.options.get("comment", [])
            comm.append("".join(words))
            c.options["comment"] = comm
            descriptor = {}
            universe_name = universe.name()
            fn = f"u{universe_name}.i"
            descriptor["universe"] = universe_name
            if transform:
                name = transform.name()
                if name is None:
                    # The transformation is anonymous, so, store its specification
                    # omitting redundant '*', TR0 words, and interleaving space tokens
                    descriptor["transform"] = transform.mcnp_words()[2:][1::2]
                else:
                    descriptor["transform"] = name
            descriptor["file"] = fn
            fill_descriptor[str(c.name())] = descriptor
            if universe_name not in already_processed_universes:
                move_universe_attribute_to_comments(universe)
                save_mcnp(universe, output / fn, override)
                logger.debug("The universe {} has been saved to {}", universe_name, fn)
                already_processed_universes.add(universe_name)

    named_transformations_descriptor = {}
    named_transformations = sorted(named_transformations, key=lambda x: x.name())
    for t in named_transformations:
        named_transformations_descriptor[f"tr{t.name()}"] = t.mcnp_words()[2:][1::2]
    fill_descriptor["named_transformations"] = named_transformations_descriptor

    fdp = output / fill_descriptor_path
    with open(fdp, "wb") as fid:
        tw.dump(fill_descriptor, fid)
    logger.debug("Fill descriptor is saved in {}", fdp)
    envelopes_path = output / "envelopes.i"
    save_mcnp(model, envelopes_path, override)
    logger.debug("The envelopes are saved to {}", envelopes_path)



from typing import Any, Dict, Iterable, Tuple, Union, cast

import math
import os
import sys

from functools import reduce
from operator import xor

import numpy as np

from .card import Card

__all__ = ["AVOGADRO", "Element", "Composition", "Material"]

AVOGADRO = 6.0221408576e23
MATERIAL_FRACTION_FORMAT = "{0:.6e}"
_CHARGE_TO_NAME = {}
_NAME_TO_CHARGE = {}
_NATURAL_ABUNDANCE = {}
_ISOTOPE_MASS = {}
_path = os.path.dirname(sys.modules[__name__].__file__)

with open(_path + "/data/isotopes.dat") as f:
    for line in f:
        number, name, *data = line.split()
        number = int(number)
        name = name.upper()
        _CHARGE_TO_NAME[number] = name
        _NAME_TO_CHARGE[name] = number
        _NATURAL_ABUNDANCE[number] = {}
        _ISOTOPE_MASS[number] = {}
        for i in range(len(data) // 3):
            isotope = int(data[i * 3])
            _ISOTOPE_MASS[number][isotope] = float(data[i * 3 + 1])
            abundance = data[i * 3 + 2]
            if abundance != "*":
                _NATURAL_ABUNDANCE[number][isotope] = float(abundance) / 100.0


TFraction = Tuple[Union["Element", int, str], float]
TFractions = Iterable[TFraction]


class Composition(Card):
    """Represents composition.

    Composition is not a material. It specifies only isotopes and their
    fractions. It doesn't concern absolute quantities like density and
    concentration. Composition immediately corresponds to the MCNP's
    material.

    weight and atomic are both lists of tuples: (element, fraction, ...).

    Parameters
    ----------
    atomic : list
        Atomic fractions.
    weight : list
        Weight fractions.
    options : dict
        Dictionary of composition options.

    Methods
    -------
    molar_mass()
        Gets molar mass of the composition.
    get_atomic(isotope)
        Gets atomic fraction of the isotope.
    get_weight(isotope)
        Gets weight fraction of the isotope.
    expand()
        Expands natural abundance elements into detailed isotope composition.
    natural(tolerance)
        Try to replace detailed isotope composition with natural elements.
    """

    _tolerance = 1.0e-3
    """Relative composition element concentration equality tolerance."""

    # TODO dvp: are there specs using both atomic and weight definitions?
    def __init__(self, atomic: TFractions = None, weight: TFractions = None, **options: Any):
        Card.__init__(self, **options)
        self._composition = {}  # type Dict[Element, float]
        elem_w = []
        frac_w = []

        if weight:
            for elem, frac in weight:
                if not isinstance(elem, Element):
                    elem = Element(elem)
                elem_w.append(elem)
                frac_w.append(frac)

        elem_a = []
        frac_a = []
        if atomic:
            for elem, frac in atomic:
                if not isinstance(elem, Element):
                    elem = Element(elem)
                elem_a.append(elem)
                frac_a.append(frac)

        if len(frac_w) + len(frac_a) > 0:
            total_frac_w = np.sum(frac_w)
            total_frac_a = np.sum(frac_a)
            atoms_in_weight_spec = np.sum(np.divide(frac_w, [e.molar_mass for e in elem_w]))
            mass_in_atomic_spec = np.sum(np.multiply(frac_a, [e.molar_mass for e in elem_a]))

            totals_diff = total_frac_a - total_frac_w
            sq_root = np.sqrt(totals_diff**2 + 4 * atoms_in_weight_spec * mass_in_atomic_spec)
            if totals_diff <= 0:
                self._mu = 0.5 * (sq_root - totals_diff) / atoms_in_weight_spec
            else:
                self._mu = 2 * mass_in_atomic_spec / (sq_root + totals_diff)

            norm_factor = self._mu * atoms_in_weight_spec + total_frac_a
            for el, frac in zip(elem_w, frac_w):
                if el not in self._composition.keys():
                    self._composition[el] = 0.0
                self._composition[el] += self._mu / norm_factor * frac / el.molar_mass
            for el, frac in zip(elem_a, frac_a):
                if el not in self._composition.keys():
                    self._composition[el] = 0.0
                self._composition[el] += frac / norm_factor
        else:
            raise ValueError("Incorrect set of parameters.")
        self._hash = reduce(xor, map(hash, self._composition.keys()))

    def copy(self):
        return Composition(atomic=cast(TFractions, self._composition.items()), **self.options)

    def __eq__(self, other):
        if len(self._composition.keys()) != len(other._composition.keys()):
            return False
        for k1, v1 in self._composition.items():
            v2 = other._composition.get(k1, None)
            if v2 is None:
                return False
            # TODO dvp: this violates exact equality definition, define separate
            #           'approx' object and method for that if exact equality is required
            if not math.isclose(v1, v2, rel_tol=Composition._tolerance):
                return False
        return True

    #
    # TODO dvp: see TODO above, implementation variant
    #
    # from dataclasses import dataclass
    # @dataclass(eq=False, frozen=True)
    # class Approx:
    #      composition: "Composition"
    #      rel_tol: float = 1e-3
    #      abs_tol: float = 1e-12
    #           ...
    #      def __hash__(self):
    #          return hash(self.composition)
    #
    #      def __eq__(self, other)
    #          ...
    # def approx(self, rel_tol=1e-3, abs_tol=1e-12):
    #     return Approx(self, rel_tol=rel_tol, abs_tol=abs_tol)

    def __hash__(self):
        return reduce(
            xor, map(hash, self._composition.keys())
        )  # TODO dvp: check why self._hash is not used?

    def mcnp_words(self, pretty=False):
        words = ["M{0} ".format(self.name())]
        for elem, frac in self._composition.items():
            words.append(elem.mcnp_repr())
            words.append("  ")
            words.append(MATERIAL_FRACTION_FORMAT.format(frac))
            words.append("\n")
        return words

    def __getitem__(self, key):
        return self.options[key]

    def __iter__(self):
        return iter(self._composition.items())

    def __contains__(self, item: Union[str, "Element"]) -> bool:
        """Checks if the composition contains the item.

        Parameters
        ----------
        item :
            Isotope. It can be either isotope name or Element instance.

        Returns
        -------
        result :
            True if the composition contains the isotope, False otherwise.
        """
        if not isinstance(item, Element):
            item = Element(item)
        return item in self._composition

    def get_atomic(self, isotope):
        """Gets atomic fraction of the isotope.

        Raises KeyError if the composition doesn't contain the isotope.

        Parameters
        ----------
        isotope : str or Element
            Isotope. It can be either isotope name or Element instance.

        Returns
        -------
        frac : float
            Atomic fraction of the specified isotope.
        """
        if not isinstance(isotope, Element):
            isotope = Element(isotope)
        return self._composition[isotope]

    def get_weight(self, isotope):
        """Gets weight fraction of the isotope.

        Raises KeyError if the composition doesn't contain the isotope.

        Parameters
        ----------
        isotope : str or Element
            Isotope. It can be either isotope name or Element instance.

        Returns
        -------
        frac : float
            Weight fraction of the specified isotope.
        """
        if not isinstance(isotope, Element):
            isotope = Element(isotope)
        at = self._composition[isotope]
        return at * isotope.molar_mass / self._mu

    @property
    def molar_mass(self):
        """Gets composition's effective molar mass [g / mol]."""
        return self._mu

    def expand(self):
        """Expands elements with natural abundances into detailed isotope composition.

        Returns
        -------
        new_comp : Composition
            New expanded composition or self.
        """
        composition = {}
        already = True
        for el, conc in self._composition.items():
            if el.mass_number == 0:
                already = False
            for isotope, frac in el.expand().items():
                if isotope not in composition.keys():
                    composition[isotope] = 0
                composition[isotope] += conc * frac
        if already:
            return self
        else:
            return Composition(atomic=composition.items(), **self.options)

    def natural(self, tolerance=1.0e-8):
        """Tries to replace detailed isotope composition by natural elements.

        Modifies current object.

        Parameters
        ----------
        tolerance : float
            Relative tolerance to consider isotope fractions as equal.
            Default: 1.e-8

        Returns
        -------
        comp : Composition
            self, if composition is reduced successfully to natural. None returned if the
            composition cannot be reduced to natural.
        """
        already = True
        by_charge = {}  # type: Dict[int, Dict[int, float]]
        for elem, fraction in self._composition.items():
            q = elem.charge
            if q not in by_charge.keys():
                by_charge[q] = {}  # type: Dict[int, float]
            a = elem.mass_number
            if a > 0:
                already = False
            if a not in by_charge[q].keys():
                by_charge[q][a] = 0
            by_charge[q][a] += fraction

        if already:  # No need for further checking - only natural elements present.
            return self

        composition = {}  # type: Dict[Element, float]
        for q, isotopes in by_charge.items():
            frac_0 = isotopes.pop(0, None)
            tot_frac = sum(isotopes.values())
            for a, fraction in isotopes.items():
                normalized_fraction = fraction / tot_frac
                delta = (
                    2
                    * abs(normalized_fraction - _NATURAL_ABUNDANCE[q][a])
                    / (normalized_fraction + _NATURAL_ABUNDANCE[q][a])
                )
                if delta > tolerance:
                    return None
            elem = Element(q * 1000)
            composition[elem] = tot_frac
            if frac_0:
                composition[elem] += frac_0
        return Composition(atomic=cast(TFractions, composition.items()), **self.options)

    def elements(self):
        """Gets iterator over composition's elements."""
        return iter(self._composition.keys())

    @staticmethod
    def mixture(*compositions: Tuple["Composition", float]) -> "Composition":
        """Makes mixture of the compositions with specific fractions.

        Parameters
        ----------
        compositions :
            List of pairs composition, fraction.

        Returns
        -------
        mix :
            Mixture.
        """
        atomics = []
        if len(compositions) == 1:
            return compositions[0][0]
        for comp, frac in compositions:
            for elem, atom_frac in comp:
                atomics.append((elem, atom_frac * frac))
        return Composition(atomic=atomics)


class Material:
    """Represents material.

    If only one of `weight` or `atomic` parameters is specified, then the Material
    there's no need to normalize it.

    Parameters
    ----------
    atomic : TFractions
        Atomic fractions. New composition will be created.
    weight : TFractions
        Weight fractions of isotopes. density of concentration must present.
    composition : Composition
        Composition instance. If it is specified, then this composition will be
        used. Neither atomic nor weight must be present.
    density : float
        Density of the material (g/cc). It is incompatible with concentration
        parameter.
    concentration : float
        Sets the atomic concentration (1 / cc). It is incompatible with density
        parameter.
    options : dict
        Extra options.

    Properties
    ----------
    density : float
        Density of the material [g/cc].
    concentration : float
        Concentration of the material [atoms/cc].
    composition : Composition
        Material's composition.
    molar_mass : float
        Material's molar mass [g/mol].

    Methods
    -------
    correct(old_vol, new_vol)
        Correct material density - returns the corrected version of the
        material.
    mixture(*materials, fraction_type)
        Makes a mixture of specified materials in desired proportions.
    __getitem__(key)
        Gets specific option.
    """

    # Relative density tolerance. Relative difference in densities when materials
    _tolerance = 1.0e-3

    def __init__(
        self,
        atomic: TFractions = None,
        weight: TFractions = None,
        composition: Composition = None,
        density: float = None,
        concentration: float = None,
        **options,
    ):
        # Attributes: _n - atomic density (concentration)
        if isinstance(composition, Composition) and not atomic and not weight:
            self._composition = composition
        elif not composition and (weight or atomic):
            self._composition = Composition(atomic=atomic, weight=weight, **options)
        else:
            raise ValueError("Incorrect set of parameters.")

        if concentration and density or not concentration and not density:
            raise ValueError("Incorrect set of parameters.")
        elif concentration:
            self._n = concentration
        else:
            self._n = density * AVOGADRO / self._composition.molar_mass
        self._options = options

    def __eq__(self, other):
        if not math.isclose(self._n, other.concentration, rel_tol=self._tolerance):
            return False
        return self._composition == other.composition

    def __hash__(self):
        return hash(self._composition)

    def __getitem__(self, key):
        return self._options[key]

    @property
    def density(self):
        """Gets material's density [g per cc]."""
        return self._n * self._composition.molar_mass / AVOGADRO

    @property
    def concentration(self):
        """Gets material's concentration [atoms per cc]."""
        return self._n

    @property
    def composition(self):
        """Gets Composition instance that corresponds to the material."""
        return self._composition

    @property
    def molar_mass(self):
        """Gets material's effective molar mass [g / mol]."""
        return self._composition.molar_mass

    def correct(self, old_vol=None, new_vol=None, factor=None):
        """Creates new material with fixed density to keep cell's mass.

        Either old_vol and new_vol or factor must be specified.

        Parameters
        ----------
        old_vol : float
            Initial volume of the cell.
        new_vol : float
            New volume of the cell.
        factor : float
            By this factor density of material will be multiplied. If factor
            is specified, then its value will be used.

        Returns
        -------
        new_mat : Material
            New material that takes into account new volume of the cell.
        """
        if factor is None:
            factor = old_vol / new_vol
        return Material(
            composition=self._composition,
            concentration=self._n * factor,
            **self._options,
        )

    @staticmethod
    def mixture(*materials, fraction_type="weight"):
        """Creates new material as a mixture of others.

        Volume fractions are not needed to be normalized, but normalization has effect.
        If the sum of fractions is less than 1, then missing fraction is considered
        to be void (density is reduced). If the sum of fractions is greater than 1,
        the effect of compression is taking place. But for weight and atomic fractions
        normalization will be done anyway.

        Parameters
        ----------
        materials : list
            A list of pairs material-fraction. material must be a Material class
            instance because for mixture not only composition but density is
            important.
        fraction_type : str
            Indicate how fraction should be interpreted.
            'weight' - weight fractions (default);
            'volume' - volume fractions;
            'atomic' - atomic fractions.

        Returns
        -------
        material : Material
            New material.
        """
        if not materials:
            raise ValueError("At least one material must be specified.")
        if fraction_type == "weight":
            fun = lambda m, f: f / m.molar_mass
            norm = sum(fun(m, f) / m.concentration for m, f in materials)
        elif fraction_type == "volume":
            fun = lambda m, f: f * m.concentration
            norm = 1
        elif fraction_type == "atomic":
            fun = lambda m, f: f
            norm = sum(fun(m, f) / m.concentration for m, f in materials)
        else:
            raise ValueError("Unknown fraction type")
        factor = sum([fun(m, f) for m, f in materials])
        compositions = [(m.composition, fun(m, f) / factor) for m, f in materials]
        new_comp = Composition.mixture(*compositions)
        return Material(composition=new_comp, concentration=factor / norm)


class Element:
    """Represents isotope or isotope mixture for natural abundance case.

    Parameters
    ----------
    _name : str or int
        Name of isotope. It can be ZAID = Z * 1000 + A, where Z - charge,
        A - the number of protons and neutrons. If A = 0, then natural abundance
        is used. Also it can be an atom_name optionally followed by '-' and A.
        '-' can be omitted. If there is no A, then A is assumed to be 0.
    comment : str, optional
        Optional comment to the element.
    lib : str
        Data library ID. Usually it is MCNP library, like '31b' for FENDL31b.
    isomer : int
        Isomer level. Default 0. Usually may appear in FISPACT output.

    Methods
    -------
    expand()
        Expands natural composition of this element.
    fispact_repr()
        Gets FISPACT representation of the element.
    mcnp_repr()
        Gets MCNP representation of the element.
    """

    def __init__(self, _name: Union[str, int], lib=None, isomer=0, comment=None):
        if isinstance(_name, int):
            self._charge = _name // 1000
            self._mass_number = _name % 1000
        else:
            z, a = self._split_name(_name.upper())
            if z.isalpha():
                self._charge = _NAME_TO_CHARGE[z]
            else:
                self._charge = int(z)
            self._mass_number = int(a)

        # molar mass calculation
        Z = self._charge
        A = self._mass_number
        if A > 0:
            if A in _ISOTOPE_MASS[Z].keys():
                self._molar = _ISOTOPE_MASS[Z][A]
            else:  # If no data about molar mass present, then mass number
                self._molar = A  # itself is the best approximation.
        else:  # natural abundance
            self._molar = 0.0
            for at_num, frac in _NATURAL_ABUNDANCE[Z].items():
                self._molar += _ISOTOPE_MASS[Z][at_num] * frac
        # Other flags and parameters
        if isinstance(lib, str):
            lib = lib.lower()
        self._lib = lib
        if self._mass_number == 0:
            isomer = 0
        self._isomer = isomer
        self._comment = comment

    def __hash__(self):
        return self._charge * (self._mass_number + 1) * (self._isomer + 1)

    def __eq__(self, other):
        if (
            self._charge == other.charge
            and self._mass_number == other.mass_number
            and self._isomer == other._isomer
        ):
            return True
        else:
            return False

    def __str__(self):
        _name = _CHARGE_TO_NAME[self.charge].capitalize()
        if self._mass_number > 0:
            _name += "-" + str(self._mass_number)
            if self._isomer > 0:
                _name += "m"
            if self._isomer > 1:
                _name += str(self._isomer - 1)
        return _name

    def mcnp_repr(self):
        """Gets MCNP representation of the element."""
        _name = str(self.charge * 1000 + self.mass_number)
        if self.lib is not None:
            _name += ".{0}".format(self.lib)
        return _name

    def fispact_repr(self):
        """Gets FISPACT representation of the element."""
        _name = _CHARGE_TO_NAME[self.charge].capitalize()
        if self._mass_number > 0:
            _name += str(self._mass_number)
            if self._isomer > 0:
                _name += "m"
            if self._isomer > 1:
                _name += str(self._isomer - 1)
        return _name

    @property
    def charge(self) -> int:
        """Gets element's charge number."""
        return self._charge

    @property
    def mass_number(self):
        """Gets element's mass number."""
        return self._mass_number

    @property
    def molar_mass(self):
        """Gets element's molar mass."""
        return self._molar

    @property
    def lib(self):
        """Gets library name."""
        return self._lib

    @property
    def isomer(self):
        """Gets isomer level."""
        return self._isomer

    def expand(self):
        """Expands natural element into individual isotopes.

        Returns
        -------
        elements : dict
            A dictionary of elements that are comprised by this one.
            Keys - elements - Element instances, values - atomic fractions.
        """
        result = {}
        if self._mass_number > 0 and self._mass_number in _NATURAL_ABUNDANCE[self._charge].keys():
            result[self] = 1.0
        elif self._mass_number == 0:
            for at_num, frac in _NATURAL_ABUNDANCE[self._charge].items():
                elem_name = "{0:d}{1:03d}".format(self._charge, at_num)
                result[Element(elem_name, lib=self._lib)] = frac
        return result

    @staticmethod
    def _split_name(_name: str) -> Tuple[str, str]:
        """Splits element's name into charge and mass number parts."""
        if _name.isnumeric():
            return _name[:-3], _name[-3:]
        for i, l in enumerate(_name):
            if l.isdigit():
                break
        else:
            return _name, "0"
        q = _name[: i - 1] if _name[i - 1] == "-" else _name[:i]
        return q, _name[i:]



from pathlib import Path

import pytest

from mckit.cli.runner import mckit
from mckit.parser import from_file
from mckit.universe import collect_transformations
from mckit.utils.resource import filename_resolver

data_filename_resolver = filename_resolver("tests.cli")


def test_help_compose(runner):
    result = runner.invoke(mckit, args=["compose", "--help"], catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "Usage: mckit compose" in result.output


def test_when_there_is_no_args(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["compose"], catch_exceptions=False)
        assert result.exit_code != 0, "Should fail when no arguments provided"
        assert "Usage:" in result.output


def test_not_existing_envelopes_file(runner):
    result = runner.invoke(mckit, args=["compose", "not-existing.imcnp"], catch_exceptions=False)
    assert result.exit_code > 0
    assert "Path 'not-existing.imcnp' does not exist" in result.output


def test_when_output_is_not_specified(runner):
    source = data_filename_resolver("data/simple_cubes.mcnp")
    result = runner.invoke(mckit, args=["compose", str(source)], catch_exceptions=False)
    assert result.exit_code > 0
    assert "Missing option '--output'" in result.output


@pytest.mark.parametrize(
    "source, output, expected",
    [
        (
            "data/simple_cubes.universes/envelopes.i",
            "simple_cubes_restored.i",
            "data/simple_cubes.mcnp",
        )
    ],
)
def test_when_fill_descriptor_is_not_specified(runner, source, output, expected):
    source = data_filename_resolver(source)
    with runner.isolated_filesystem() as test_folder:
        result = runner.invoke(
            mckit, args=["compose", "--output", output, source], catch_exceptions=False
        )
        assert (
            result.exit_code == 0
        ), "Should success using fill_descriptor in the same directory as source file"
        assert Path(output).exists(), f"Should create file {output} file in {test_folder}"
        actual = from_file(output)
        expected = from_file(data_filename_resolver(expected))
        assert actual.universe.has_equivalent_cells(expected.universe), "Cells differ"


@pytest.mark.parametrize(
    "source, output, expected",
    [
        (
            "data/cubes_with_fill_transforms.universes/envelopes.i",
            "cubes_with_fill_transforms.i",
            "data/cubes_with_fill_transforms.mcnp",
        )
    ],
)
def test_anonymous_transforms(runner, source, output, expected):
    source = data_filename_resolver(source)
    with runner.isolated_filesystem() as test_folder:
        result = runner.invoke(
            mckit, args=["compose", "--output", output, source], catch_exceptions=False
        )
        assert (
            result.exit_code == 0
        ), "Should success using fill_descriptor in the same directory as source file"
        assert Path(output).exists(), f"Should create file {output} file in {test_folder}"
        actual = from_file(output)
        expected = from_file(data_filename_resolver(expected))
        assert actual.universe.has_equivalent_cells(expected.universe), "Cells differ"


@pytest.mark.parametrize(
    "universes",
    [
        "cubes_with_fill_named_transforms",
        "two_cubes_with_the_same_filler",
        "shared_surface",
    ],
)
def test_compose(runner, universes):
    source = f"data/{universes}.universes/envelopes.i"
    output = f"{universes}.i"
    expected = f"data/{universes}.mcnp"
    source = data_filename_resolver(source)
    assert Path(source).exists(), f"File {source} does not exist"
    with runner.isolated_filesystem() as test_folder:
        result = runner.invoke(
            mckit, args=["compose", "--output", output, source], catch_exceptions=False
        )
        assert (
            result.exit_code == 0
        ), "Should success using fill_descriptor in the same directory as source file"
        assert Path(output).exists(), f"Should create file {output} file in {test_folder}"
        actual = from_file(output)
        expected = from_file(data_filename_resolver(expected))
        assert actual.universe.has_equivalent_cells(expected.universe), "Cells differ"
        actual_transformations = collect_transformations(actual.universe)
        expected_transformations = collect_transformations(expected.universe)
        assert (
            actual_transformations == expected_transformations
        ), "The transformations should be the same"



import mckit.parser.common.utils as cmn
import sly

from mckit.material import Composition, Element
from mckit.parser.common.Lexer import Lexer as LexerBase
from mckit.parser.common.utils import drop_comments


# noinspection PyPep8Naming,PyUnboundLocalVariable,PyUnresolvedReferences,SpellCheckingInspection
class Lexer(LexerBase):
    tokens = {NAME, FRACTION, OPTION}
    ignore = " \t="
    OPTION = r"(?:(?:gas|estep|cond)\s*[ =]\s*\d+)|(?:(?:n|p|pn|e)lib\s*[ =]\s*\S+)"

    @_(r"m\d+")
    def NAME(self, t):
        t.value = int(t.value[1:])
        return t

    @_(r"\d+(?:\.\d+[cdepuy])?\s+[-+]?((\d+(\.\d*)?)|(\.\d+))([eE][-+]?\d+)?")
    def FRACTION(self, t):
        isotop_spec, frac = t.value.split()
        frac = float(frac)
        if "." in isotop_spec:
            isotope, lib = isotop_spec.split(".")
            isotope = int(isotope)
            lib = cmn.ensure_lower(lib)
            t.value = isotope, lib, frac
        else:
            isotope = int(isotop_spec)
            t.value = isotope, None, frac
        return t


# noinspection PyUnresolvedReferences
class Parser(sly.Parser):
    tokens = Lexer.tokens

    def build_composition(self, name, fractions, options=None):
        atomic = []
        weight = []

        for el, fraction in fractions:
            if fraction < 0.0:
                weight.append((el, -fraction))
            else:
                atomic.append((el, fraction))

        if options is None:
            options = {}

        options["name"] = name

        return Composition(atomic=atomic, weight=weight, **options)

    @_("composition_a")
    def composition(self, p):
        name, fractions, options = p.composition_a
        return self.build_composition(name, fractions, options)

    @_("NAME fractions options")
    def composition_a(self, p):
        return p.NAME, p.fractions, p.options

    @_("NAME fractions")
    def composition_a(self, p):
        return p.NAME, p.fractions, None

    @_("fractions fraction")
    def fractions(self, p):
        p.fractions.append(p.fraction)
        return p.fractions

    @_("fraction")
    def fractions(self, p):
        return [p.fraction]

    # @_('fraction_a EOL_COMMENT')
    # def fraction(self, p):
    #     name, lib, frac = p.fraction_a
    #     return Element(name, lib=lib, comment=p.EOL_COMMENT), frac
    #
    @_("FRACTION")
    def fraction(self, p):
        name, lib, frac = p.FRACTION
        return Element(name, lib=lib), frac

    @_("options option")
    def options(self, p):
        option, value = p.option
        p.options[option] = value
        return p.options

    @_("option")
    def options(self, p):
        option, value = p.option
        return {option: value}

    @_("OPTION")
    def option(self, p):
        text: str = p.OPTION
        if "=" in text:
            text = text.replace("=", " ", 1)
        option, value = text.split()
        option = cmn.ensure_upper(option)
        if option in ("GAS", "ESTEP", "COND"):
            value = int(value)
        return option, value


def parse(text):
    text = drop_comments(text)
    lexer = Lexer()
    parser = Parser()
    result = parser.parse(lexer.tokenize(text))
    return result



from typing import Any

from contextlib import contextmanager

from mckit.utils.accept import accept, on_unknown_acceptor


def test_accept_single_level():
    class DemoContainer(list):
        pass

    class DemoItem:
        def __init__(self, a: int):
            self.name = a

    dc = DemoContainer(map(DemoItem, range(10)))
    my_sum = sum(range(10))

    @contextmanager
    def visit_container(demo_obj: Any) -> int:
        if isinstance(demo_obj, DemoContainer):

            def at_item(a: Any, b: DemoItem) -> Any:
                a[0] += b.name
                return a

            yield at_item, [0, len(demo_obj)]
        else:
            on_unknown_acceptor(demo_obj)

    _sum, _len = accept(dc, visit_container)
    assert _sum == my_sum
    assert _len == 10


def test_accept_two_levels():
    class DemoContainer(list):
        pass

    class Intermediate(list):
        pass

    class DemoItem:
        def __init__(self, a: int):
            self.name = a

        def __repr__(self):
            return f"Item({self.name})"

    dc = DemoContainer(map(Intermediate, [map(DemoItem, range(2)), map(DemoItem, range(2, 4))]))
    expected = [[sum(range(2)), 2], [sum(range(2, 4)), 2]]

    def at_item(initial: Any, b: DemoItem) -> Any:
        initial[0] += b.name
        return initial

    @contextmanager
    def visit_intermediate(intermediate: Intermediate):
        if isinstance(intermediate, Intermediate):
            initial_for_item = [0, len(intermediate)]
            yield at_item, initial_for_item
        else:
            on_unknown_acceptor(intermediate)

    def at_intermediate(initial: Any, intermediate: Intermediate):
        initial.append(accept(intermediate, visit_intermediate))
        return initial

    @contextmanager
    def visit_container(demo_obj: Any) -> Any:
        if isinstance(demo_obj, DemoContainer):
            initial_for_intermediate = []
            yield at_intermediate, initial_for_intermediate
        else:
            on_unknown_acceptor(demo_obj)

    actual = accept(dc, visit_container)
    assert actual == expected



import pytest

from mckit.cli.runner import mckit
from mckit.utils.resource import filename_resolver

data_filename_resolver = filename_resolver("tests")


def test_when_there_is_no_args(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["check"], catch_exceptions=False)
        assert result.exit_code != 0, "Should fail when no arguments provided"
        assert "Usage:" in result.output


def test_not_existing_mcnp_file(runner):
    result = runner.invoke(mckit, args=["check", "not-existing.imcnp"], catch_exceptions=False)
    assert result.exit_code > 0
    assert "Path 'not-existing.imcnp' does not exist" in result.output


@pytest.mark.parametrize(
    "source, expected",
    [("cli/data/simple_cubes.mcnp", "cells;surfaces;transformations;compositions")],
)
def test_good_path(runner, source, expected):
    source = data_filename_resolver(source)
    result = runner.invoke(mckit, args=["--quiet", "check", source], catch_exceptions=False)
    assert result.exit_code == 0, "Should success"
    for e in expected.split(";"):
        assert e in result.output



# -*- coding: utf-8 -*-
from typing import Dict, List, NoReturn

import pytest

from mckit.utils.indexes import Index


def dummy_strategy(c: int) -> int:
    return c * c


@pytest.mark.parametrize("inp", [{1: 1, 2: 4}, {}])
def test_index_with_dummy_strategy(inp: Dict[int, int]) -> None:
    dictionary = Index(dummy_strategy)
    dictionary.update(inp)
    for k in range(5):
        v = dictionary[k]
        expected = k * k
        assert expected == v


class MyKeyError(KeyError):
    def __init__(self, key: int, *args, **kwargs):
        super().__init__(args, kwargs)
        self._key = key

    @property
    def key(self):
        return self._key


def strict_strategy(c: int) -> NoReturn:
    raise MyKeyError(c)


@pytest.mark.parametrize(
    "inp,keys,success",
    [({1: 1, 2: 4}, [1, 2, 3], [True, True, False]), ({}, [1], [False])],
)
def test_index_with_strict_strategy(
    inp: Dict[int, int], keys: List[int], success: List[bool]
) -> None:
    dictionary = Index(strict_strategy)
    dictionary.update(inp)
    for k, s in zip(keys, success):
        try:
            _ = dictionary[k]
            assert s is True
        except MyKeyError as ex:
            assert ex.key == k
            assert s is False



"""Classes to index MCNP objects on model file parsing."""
from __future__ import annotations

from typing import Callable, Dict, Iterable, NoReturn, Optional, Type, TypeVar, cast

from functools import reduce

from mckit.utils.named import Name, default_name_key

Key = TypeVar("Key")
Item = TypeVar("Item")
FactoryMethodWithKey = Callable[[Key], Optional[Item]]


class Index(Dict[Key, Item]):
    """Like collections.defaultdict but the factory takes key as argument.

    The class redefines __missing__ method to use `key` on calling factory method.

    Attrs:
        _default_factory (FactoryMethodWithKey):
            Same as in Args.
    """

    def __init__(
        self,
        default_factory: FactoryMethodWithKey = None,
        **kwargs: Dict[Key, Item],
    ) -> None:
        """Create `Index`.

        Args:
            default_factory: factory method accepting `key` as the only positional argument.
            kwargs: keyword arguments to pass to base class, if any
        """
        super().__init__(self, **kwargs)
        self._default_factory = default_factory

    @property
    def default_factory(self) -> FactoryMethodWithKey:
        """Public accessor to `self._default_factory`."""
        return self._default_factory

    def __missing__(self, key: Key) -> Optional[Item]:
        """Calls default factory with the key as argument."""
        if self._default_factory:
            return self._default_factory(key)

        raise NumberedItemNotFoundError(key)


# noinspection PyUnusedLocal
def ignore(_: Key) -> Optional[Item]:
    """Default factory for `IgnoringIndex`.

    Returns:
        None - always.
    """
    return None


class IgnoringIndex(Index[Key, Item]):
    """Index ignoring absence of a key in the dictionary."""

    def __init__(self, **kwargs: Dict[Key, Item]) -> None:
        Index.__init__(self, ignore, **kwargs)


class NumberedItemNotFoundError(KeyError):
    """Error to raise, when an item is not found in an `Index`."""

    kind: str = ""

    def __init__(self, key: Key) -> None:
        self._key = key
        super().__init__(f"{self.kind} #{self._key} is not found")


class NumberedItemDuplicateError(ValueError):
    """Error to raise, when an item has duplicate in an `Index`."""

    kind: str = ""

    def __init__(self, key: Key, prev: Item, curr: Item) -> None:
        self._key = key
        self._prev = prev
        self._curr = curr
        super().__init__(
            f"{self.kind} #{self._key} is duplicated, see {self._prev} and {self._curr}"
        )


def raise_on_duplicate_strategy(key: Key, prev: Item, curr: Item) -> NoReturn:
    """Raise error on `key` duplicate found, regardless values.

    Args:
        key: a key in `Index`.
        prev: the value already in `Index`.
        curr: the new value to add to `Index`.

    Raises:
        NumberedItemDuplicateError: exception to inform on `key`, `prev` and `curr` values.
    """
    raise NumberedItemDuplicateError(key, prev, curr)


def ignore_equal_objects_strategy(key: Key, prev: Item, curr: Item) -> None:
    """Raise error on `key` duplicate found, if the values are not equal.

    Otherwise, ignore an attempt to add the same key/value pair.

    Args:
        key: a key in `Index`.
        prev: the value already in `Index`.
        curr: the new value to add to `Index`.

    Raises:
        NumberedItemDuplicateError: exception to inform on `key`, `prev` and `curr` values.
    """
    if prev is not curr and prev != curr:
        raise NumberedItemDuplicateError(key, prev, curr)


class StatisticsCollector(Dict[Key, int]):
    """Duplicates counter."""

    def __init__(self, ignore_equal=False):
        super().__init__(self)
        self.ignore_equal = ignore_equal

    def __missing__(self, key):
        return 1

    def __call__(self, key: Key, prev: Item, curr: Item) -> None:
        """Increase counter.

        If items are duplicates and `self.ignore_equal` is set on,
        then skip.

        Args:
            key: a key in `StatisticsCollector`.
            prev: the value already in `StatisticsCollector`.
            curr: the new value to add to `StatisticsCollector`.
        """
        if not self.ignore_equal or prev != curr:
            self[key] += 1


class IndexOfNamed(Index[Key, Item]):
    """Index of items from a key can be extracted."""

    @classmethod
    def from_iterable(
        cls: Type["IndexOfNamed[Key, Item]"],
        items: Iterable[Item],
        *,
        key: Callable[[Item], Name] = default_name_key,
        default_factory: Callable[[Key], Item] = None,
        on_duplicate: Callable[[Key, Item, Item], None] = None,
    ) -> "IndexOfNamed":
        """Construct `IndexOfNamed` from `items`."""
        index = cls(default_factory)

        def _reducer(a, b):
            name = key(b)
            if on_duplicate is not None and name in a:
                on_duplicate(name, a[name], b)
            a[name] = b
            return a

        return cast("IndexOfNamed[Key, Item]", reduce(_reducer, items, index))



"""Shared constants."""
import numpy as np

__all__ = [
    "MIN_BOX_VOLUME",
    "RESOLUTION",
    "FLOAT_TOLERANCE",
    "MCNP_ENCODING",
    "DROP_OPTIONS",
]

MIN_BOX_VOLUME = 0.001

# Resolution of float number
RESOLUTION = np.finfo(float).resolution

FLOAT_TOLERANCE = 1.0e-12

MCNP_ENCODING = "cp1251"

DROP_OPTIONS = frozenset(["original", "transform", "comment", "trailing_comment", "comment_above"])



from typing import List, Union

import shutil
import sys
import sysconfig

from pathlib import Path

WIN = sys.platform.startswith("win32") and "mingw" not in sysconfig.get_platform()
MACOS = sys.platform.startswith("darwin")


def get_library_dir(check: bool = False) -> Path:
    root_prefix = Path(sys.prefix)

    if WIN:
        root_prefix = root_prefix / "Library"

    library_dir = root_prefix / "lib"

    if check:
        if not library_dir.is_dir():
            raise EnvironmentError(f"{library_dir} is not a valid directory")

    return library_dir


def create_directory(path: Path, clean: bool = True) -> Path:
    if clean and path.exists():
        shutil.rmtree(path)
    path.mkdir(exist_ok=True, parents=True)
    return path


def check_directories(*directories: str) -> None:
    for directory in directories:
        if not isinstance(directory, str):
            raise TypeError(f"The value {directory} is not a string")
        if not Path(directory).is_dir():
            raise EnvironmentError(f"The directory {directory} does not exist")


def insert_directories(destination: List[str], value: Union[str, List[str]]) -> List[str]:
    dirs = value
    if not isinstance(value, list):
        dirs = [dirs]
    for old in destination:
        if old not in dirs:
            dirs.append(old)
    return dirs



# -*- coding: utf-8 -*-
"""Проверяет корректность модели и выдает статистику."""
from typing import Any, Callable, Iterable, Optional

from pathlib import Path

from mckit import Universe
from mckit.card import Card
from mckit.cli.logging import logger
from mckit.parser.mcnp_input_sly_parser import ParseResult, from_file
from mckit.universe import collect_transformations


def check_duplicates(
    iterable: Optional[Iterable[Any]], label: str, key: Callable[[Any], Any]
) -> None:
    if iterable is None:
        print(f"No {label}s are found")  # don't use logger here, should go to stdout
    else:
        visited = set()
        for c in iterable:
            k = key(c)
            if k in visited:
                print(f"Duplicate of {label} {k} is found")
            else:
                visited.add(k)
        print(f"Total of {label}s: {len(visited)}")


def check(source):
    result = 0
    logger.info("Check model {}", source)
    source = Path(source)
    parse_result: ParseResult = from_file(source)
    model = parse_result.universe
    logger.debug("Read the model okay")
    universes = model.get_universes()
    check_duplicates(universes, "universe", Universe.name)
    cells = []
    surfaces = []
    transformations = []
    compositions = []
    for u in universes:
        cells.extend(u)
        surfaces.extend(u.get_surfaces(inner=False))
        transformations_to_add = collect_transformations(u)
        if transformations_to_add:
            transformations.extend(transformations_to_add)
        compositions.extend(u.get_compositions())
    if not transformations:
        transformations = None
    if not compositions:
        compositions = None
    check_duplicates(cells, "cell", Card.name)
    check_duplicates(surfaces, "surface", Card.name)
    check_duplicates(transformations, "transformation", Card.name)
    check_duplicates(compositions, "composition", Card.name)
    return result



from pathlib import Path

import pytest

from click.testing import CliRunner
from mckit.cli.commands.common import get_default_output_directory
from mckit.cli.runner import mckit
from mckit.parser.mcnp_section_parser import is_comment
from mckit.utils.io import MCNP_ENCODING
from mckit.utils.resource import filename_resolver


@pytest.fixture
def runner():
    return CliRunner()


data_filename_resolver = filename_resolver("tests")


def test_when_there_is_no_args(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["split"], catch_exceptions=False)
        assert result.exit_code != 0, "Should fail when no arguments provided"
        assert "Usage:" in result.output


def test_not_existing_mcnp_file(runner):
    result = runner.invoke(mckit, args=["split", "not-existing.imcnp"], catch_exceptions=False)
    assert result.exit_code > 0
    assert "Path 'not-existing.imcnp' does not exist" in result.output


@pytest.mark.parametrize(
    "source, out, expected",
    [
        (
            "cli/data/simple_cubes_with_tallies.mcnp",
            "somewhere.dir",
            "title.txt cells.txt surfaces.txt materials.txt tallies.txt cards.txt",
        )
    ],
)
def test_when_output_dir_is_specified(runner, source, out, expected):
    source = data_filename_resolver(source)
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["split", "-o", out, source], catch_exceptions=False)
        assert result.exit_code == 0, "Should success without output directory"
        out = Path(out)
        assert out.is_dir()
        expected = expected.split()
        for e in expected:
            assert (out / e).exists()


@pytest.mark.parametrize(
    "source, expected",
    [
        (
            "cli/data/simple_cubes.mcnp",
            "title.txt cells.txt surfaces.txt materials.txt cards.txt",
        )
    ],
)
def test_when_output_dir_is_not_specified(runner, source, expected):
    source = data_filename_resolver(source)
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["split", source], catch_exceptions=False)
        assert result.exit_code == 0, "Should success without output directory"
        out = get_default_output_directory(source, ".split")
        assert out.is_dir()
        expected = expected.split()
        for e in expected:
            assert (out / e).exists()


@pytest.mark.parametrize(
    "source, expected",
    [
        (
            "cli/data/simple_cubes.mcnp",
            "title.txt \
        cells_start.txt cells.txt cells_end.txt \
        surfaces_start.txt surfaces.txt surfaces_end.txt \
        materials_start.txt materials.txt materials_end.txt \
        transformations_start.txt transformations_end.txt \
        cards.txt \
        new_line.txt",
        )
    ],
)
def test_when_separator_files_are_required(runner, source, expected):
    source = data_filename_resolver(source)
    with runner.isolated_filesystem():
        result = runner.invoke(
            mckit, args=["split", "--separators", source], catch_exceptions=False
        )
        assert result.exit_code == 0, "Should success"
        out = get_default_output_directory(source, ".split")
        assert out.is_dir()
        expected = expected.split()
        for e in expected:
            assert (out / e).exists()
        text = (out / "cells_start.txt").read_text(encoding=MCNP_ENCODING)
        assert is_comment(text), "Should be MCNP comment text"


# @pytest.mark.parametrize("source,expected", [
#     ("cli/data/simple_cubes.mcnp", "envelopes.i u1.i u2.i".split()),
# ])
# def test_when_only_source_is_specified(runner, source, expected):
#     source: Path = data_filename_resolver(source)
#     with runner.isolated_filesystem():
#         result = runner.invoke(mckit, args=['split', source], catch_exceptions=False)
#         assert result.exit_code == 0, \
#             "Should success without specified output: " + result.output
#         output: Path = get_default_output_directory(source)
#         for f in expected:
#             p = output / f
#             assert p.exists(), \
#                 f"Should store the file {p} in the default directory '{output}'"
#             model = mk.read_mcnp(p)
#             for cell in model:
#                 assert 'U' not in cell.options or cell.options['U'].name() == 0
#
#
# @pytest.mark.parametrize("source,output,expected", [
#     ("cli/data/simple_cubes.mcnp", "split-1", "envelopes.i u1.i u2.i".split()),
# ])
# def test_when_output_is_specified(runner, source, output, expected):
#     source = data_filename_resolver(source)
#     with runner.isolated_filesystem():
#         output = Path(output)
#         assert not output.exists(), f"The {output} directory should not exist before the test run"
#         result = runner.invoke(mckit, args=['split', "--output", str(output), source], catch_exceptions=False)
#         assert result.exit_code == 0
#         assert output.exists(), f"The {output} directory should exist after the test run"
#         for f in expected:
#             assert (output / f).exists(), \
#                 f"Should store the file {f} in the default directory 'universes'"
#
#
# def test_when_output_file_exists_and_override_is_not_specified(runner):
#     source = data_filename_resolver("cli/data/simple_cubes.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         output = Path(prefix) / "universes/envelopes.i"
#         output.parent.mkdir(parents=True)
#         output.touch(exist_ok=False)
#         result = runner.invoke(
#             mckit,
#             args=["split", source],
#             catch_exceptions=False
#         )
#         assert result.exit_code != 0, \
#             "Should fail when output file exists and --override is not specified"
#
#
# def test_when_output_file_exists_and_override_is_specified(runner):
#     source = data_filename_resolver("cli/data/simple_cubes.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         output = Path(prefix) / "simple_cubes./envelopes.i"
#         output.parent.mkdir(parents=True)
#         output.touch(exist_ok=False)
#         result = runner.invoke(
#             mckit,
#             args=["--override", "split", source],
#             catch_exceptions=False
#         )
#         assert result.exit_code == 0, \
#             "Should success when output file exists and --override is specified"
#
#
#
# def test_fill_descriptor(runner):
#     source = data_filename_resolver("cli/data/simple_cubes.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         output = Path(prefix) / "universes/fill-descriptor.toml"
#         result = runner.invoke(
#             mckit,
#             args=["split", source],
#             catch_exceptions=False
#         )
#         assert result.exit_code == 0, "Should success"
#         assert output.exists()
#         with output.open() as fid:
#             fill_descriptor = fid.read()
#             assert fill_descriptor.find('simple_cubes.mcnp')
#             fill_descriptor = tk.parse(fill_descriptor)
#             assert 'created' in fill_descriptor
#             assert '2' in fill_descriptor
#             assert 'universe' in fill_descriptor['2']
#             assert 1 == fill_descriptor['2']['universe']
#             assert 'u1.i' == fill_descriptor['2']['file']
#
#
#
# def test_fill_descriptor_when_fill_descriptor_file_is_specified(runner):
#     source = data_filename_resolver("cli/data/simple_cubes.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         fill_descriptor_path = Path(prefix) / "fill-descriptor-special.toml"
#         result = runner.invoke(
#             mckit,
#             args=["split", "--fill-descriptor", str(fill_descriptor_path), source],
#             catch_exceptions=False
#         )
#         assert result.exit_code == 0, "Should success"
#         assert fill_descriptor_path.exists()
#
#
# def test_anonymous_transformation(runner):
#     source = data_filename_resolver("cli/data/cubes_with_fill_transforms.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         output = Path(prefix) / "cubes_with_fill_transforms.universes"
#         result = runner.invoke(
#             mckit,
#             args=["split", source],
#             catch_exceptions=False
#         )
#         assert result.exit_code == 0, "Should success"
#         with open(output / "fill-descriptor.toml") as fid:
#             descriptor = tk.parse(fid.read())
#             spec = descriptor['2']['transform']
#             assert float(spec[1]) == -1.0, "Fill descriptor is wrong"



"""Functions for MCNP model text printing."""
from typing import Any, List

import warnings

import mckit.constants as constants

from mckit.utils import get_decades, prettify_float, significant_digits

IMPORTANCE_FORMAT = "{0:.3f}"


def print_card(tokens: List[str], offset: int = 8, max_column: int = 80, sep: str = "\n") -> str:
    """Produce string in MCNP card format.

    Parameters
    ----------
    tokens :
        List of words to be printed.
    offset :
        The number of spaces to make continuation of line. Minimum 5.
    max_column :
        The maximum length of card line. Maximum 80.
    sep :
        Separator symbol. This symbol marks positions where newline character
        should be inserted even if max_column position not reached.

    Returns
    -------
    text :
        MCNP code of a card.
    """
    if offset < 5:
        offset = 5
        warnings.warn("offset must not be less than 5. offset is set to be 5.")
    if max_column > 80:
        max_column = 80
        warnings.warn("max_column must not be greater than 80. It is set to be 80.")

    length = 0  # current length.
    words = []  # a list of individual words.
    line_sep = "\n" + " " * offset  # separator between lines.
    i = 0
    while i < len(tokens):
        if length + len(tokens[i]) > max_column or tokens[i] == sep:
            words.append(line_sep)
            length = offset
            while tokens[i] == sep or tokens[i].isspace():
                i += 1
                if i == len(tokens):
                    words.pop()
                    return "".join(words)
        words.append(tokens[i])
        length += len(tokens[i])
        i += 1
    return "".join(words)


def separate(tokens: List[str], sep: str = " ") -> List[str]:
    """Adds separation symbols between tokens.

    Parameters
    ----------
    tokens :
        A list of strings.
    sep :
        Separator to be inserted between tokens. Default: single space.

    Returns
    -------
    sep_tokens :
        List of separated tokens.
    """
    sep_tokens = []
    for t in tokens[:-1]:
        sep_tokens.append(t)
        sep_tokens.append(sep)
    sep_tokens.append(tokens[-1])
    return sep_tokens


def print_option(option: str, value: Any) -> List[str]:
    name = option[:3]
    par = option[3:]
    if name == "IMP" and (par == "N" or par == "P" or par == "E"):
        return ["IMP:{0}={1}".format(par, IMPORTANCE_FORMAT.format(value))]
    elif option == "VOL":
        return ["VOL={0}".format(value)]
    elif option == "U":
        return ["U={0}".format(value.name())]
    elif option == "FILL":
        universe = value["universe"]
        tr = value.get("transform", None)
        words = ["FILL={0}".format(universe.name())]
        if tr:
            tr_name = tr.name()
            if not tr_name:
                words[0] = "*" + words[0]
                words.append("(")
                words.extend(tr.get_words())
                words.append(")")
            else:
                words.append("(")
                words.append(str(tr_name))
                words.append(")")
        return words
    else:
        raise ValueError("Incorrect option name: {0}".format(option))


def pretty_float(value: float, frac_digits: int = None) -> str:
    """Pretty print of the float number.

    Parameters
    ----------
    value :
        Value to be printed.
    frac_digits :
        The number of digits after decimal point.
    """
    if frac_digits is None:
        frac_digits = significant_digits(
            value, constants.FLOAT_TOLERANCE, resolution=constants.FLOAT_TOLERANCE
        )
    if value == abs(value):
        value = abs(value)
    decades = get_decades(value)
    format_f = "{{0:.{0}f}}".format(max(frac_digits, 0))
    format_e = "{{0:.{0}e}}".format(max(frac_digits + decades, 0))
    text_f = format_f.format(round(value, frac_digits))
    text_e = format_e.format(value)
    if len(text_f) <= len(text_e):
        return text_f
    else:
        return text_e


CELL_OPTION_GROUPS = (
    ("IMPN", "IMPP", "IMPE", "VOL"),  # Importance options
    ("TRCL",),  # Transformation options
    ("U", "FILL"),  # Universe and fill options
)


def add_float(words: List[str], v: float, pretty: bool) -> None:
    words.append(" ")
    if pretty:
        words.append(pretty_float(v))
    else:
        words.append(prettify_float(v))



from __future__ import annotations

import typing as tp

from typing import Iterable, List, NewType, Optional, Set, Union

import os

from copy import deepcopy
from functools import reduce
from itertools import groupby, permutations, product
from multiprocessing import Pool

import mckit.material as mm
import numpy as np

from click import progressbar

# noinspection PyUnresolvedReferences,PyPackageRequirements
from mckit.geometry import Shape as _Shape

from .box import GLOBAL_BOX, Box
from .card import Card
from .constants import MIN_BOX_VOLUME
from .printer import CELL_OPTION_GROUPS, print_option
from .surface import Surface
from .transformation import Transformation
from .utils import filter_dict

__all__ = ["Shape", "Body", "simplify", "GLOBAL_BOX", "Card", "TGeometry", "TGeometry"]


# noinspection PyProtectedMember
class Shape(_Shape):
    """Describes shape.

    Note:
        Shape is immutable object.

    Args:
        opc:
            Operation code. Denotes operation to be applied. Possible values:
            'I' - for intersection;
            'U' - for union;
            'C' - for complement;
            'S' - (same) no operation;
            'E' - empty set - no space occupied;
            'R' - whole space.
        args:
            Geometry elements. It can be either Shape or Surface instances. But
            no arguments must be specified for 'E' or 'R' opc. Only one argument
            must present for 'C' or 'S' opc values.

    Attrs:
        opc (str):
            Operation code. It may be different from opc passed in __init__.
        invert_opc (str):
            Operation code, complement to the opc.
        args (Tuple[Shape|Surface...]):
            A tuple of shape's arguments.

    Methods (inherited from native parent):
        test_box(box)
            Tests if the box intersects the shape.
        volume(box, min_volume)
            Calculates the volume of the shape with desired accuracy.
        bounding_box(box, tol)
            Finds bounding box for the shape with desired accuracy.
        test_points(points)
            Tests the senses of the points.
        is_complement(other)
            Checks if other is a complement to the shape.
        complement()
            Gets complement shape.
        union(*other)
            Creates a union of the shape with the others.
        intersection(*other)
            Creates an intersection of the shape with the others.
        transform(tr)
            Gets transformed version of the shape.
        is_empty()
            Checks if this shape is empty - no space belong to it.
        get_surfaces()
            Gets all Surface objects that bounds the shape.
        complexity()
            Gets the complexity of the shape description.
        get_simplest()
            Gets the simplest description of the shape.
        replace_surfaces(replace_dict)
            Creates new Shape object by replacing surfaces.
    """

    _opc_hash = {
        "I": hash("I"),
        "U": ~hash("I"),
        "E": hash("E"),
        "R": ~hash("E"),
        "S": hash("S"),
        "C": ~hash("S"),
    }

    def __init__(self, opc: str, *args: Union["Shape", "Surface"]):
        opc, args = _clean_args(opc, *args)
        _Shape.__init__(self, opc, *args)
        self._calculate_hash(opc, *args)

    def __iter__(self):
        return iter(self.args)

    def __getstate__(self):
        return self.opc, self.args, self._hash

    def __setstate__(self, state):
        opc, args, hash_value = state
        _Shape.__init__(self, opc, *args)
        self._hash = hash_value

    def __repr__(self):
        return f"Shape({self.opc}, {self.args})"

    def _get_words(self, parent_opc: str = None) -> List[str]:
        """Gets list of words that describe the shape.

        Args:
            parent_opc:  Operation code of parent shape.
                         It is needed for proper use of parenthesis.

        Returns:
            List of words.
        """
        words = []
        if self.opc == "S":
            words.append(f"{self.args[0].options['name']}")
        elif self.opc == "C":
            words.append(f"-{self.args[0].options['name']}")
        elif self.opc == "E":
            words.append("EMPTY_SET")
        elif self.opc == "R":
            words.append("UNIVERSE_SET")
        else:
            sep = " " if self.opc == "I" else ":"
            args = self.args
            need_parentheses = self.opc == "U" and parent_opc == "I"
            if need_parentheses:
                words.append("(")
            for a in args[:-1]:
                words.extend(a._get_words(self.opc))
                words.append(sep)
            words.extend(args[-1]._get_words(self.opc))
            if need_parentheses:
                words.append(")")
        return words

    def __eq__(self, other):
        if self is other:
            return True
        if self.opc != other.opc:
            return False
        if self.opc == "E" or self.opc == "R":
            return True
        if len(self.args) != len(other.args):
            return False
        self_groups = {k: list(v) for k, v in groupby(sorted(self.args, key=hash), key=hash)}
        other_groups = {k: list(v) for k, v in groupby(sorted(other.args, key=hash), key=hash)}
        flag = False  # TODO dvp: check is this statement doesn't break tests
        for hash_value, entities in self_groups.items():
            flag = False
            if hash_value not in other_groups.keys():
                return False
            if len(entities) != len(other_groups[hash_value]):
                return False
            for other_entities in permutations(other_groups[hash_value]):
                for se, oe in zip(entities, other_entities):
                    if not (se == oe):
                        break
                else:
                    flag = True
                    break
            if not flag:
                return False

        if flag:
            return True
        return False

    def __hash__(self):
        return self._hash

    def _calculate_hash(self, opc, *args):
        """Calculates hash value for the object.

        Hash is 'xor' for hash values of all arguments together with opc hash.
        """
        if opc == "C":  # C and S can be present only with Surface instance.
            self._hash = ~hash(args[0])
        elif opc == "S":
            self._hash = hash(args[0])
        else:
            self._hash = self._opc_hash[opc]
            for a in args:
                self._hash ^= hash(a)

    def complement(self):
        """Gets complement to the shape.

        Returns
        -------
        comp_shape : Shape
            Complement shape.
        """
        opc = self.opc
        args = self.args
        if opc == "S":
            return Shape("C", args[0])
        elif opc == "C":
            return Shape("S", args[0])
        elif opc == "E":
            return Shape("R")
        elif opc == "R":
            return Shape("E")
        else:
            opc = self.invert_opc
            c_args = [a.complement() for a in args]
            return Shape(opc, *c_args)

    def is_complement(self, other):
        """Checks if this shape is complement to the other.

        Returns
        -------
        result : bool
            Test result.
        """
        if hash(self) != ~hash(other):
            return False
        if self.opc != other.invert_opc:
            return False
        if len(self.args) != len(other.args):
            return False
        if len(self.args) == 1:
            return self.args[0] == other.args[0]
        elif len(self.args) > 1:
            for a in self.args:
                for b in other.args:
                    if a.is_complement(b):
                        break
                else:
                    return False
        return True

    def intersection(self, *other: Union["Shape", "Body"]) -> "Shape":
        """Gets intersection with other shape.

        Parameters
        ----------
        other : tuple
            A list of Shape or Body objects, which must be intersected.

        Returns
        -------
        result : Shape
            New shape.
        """
        return Shape("I", self, *other)

    def union(self, *other):
        """Gets union with other shape.

        Parameters
        ----------
        other : tuple
            A list of Shape or Body objects, which must be joined.

        Returns
        -------
        result : Shape
            New shape.
        """
        return Shape("U", self, *other)

    def transform(self, tr):
        """Transforms the shape.

        Parameters
        ----------
        tr : Transformation
            Transformation to be applied.

        Returns
        -------
        result : Shape
            New shape.
        """
        opc = self.opc
        args = []
        for a in self.args:
            a = a.transform(tr)
            if isinstance(a, Surface):
                a = a.apply_transformation()
                # TODO dvp: check if call of apply_transformation() should be moved to caller site
                #           it would be better to change only transformations instead of the surfaces
            args.append(a)
        return Shape(opc, *args)

    def apply_transformation(self):
        opc = self.opc
        args = []
        for a in self.args:
            if hasattr(a, "apply_transformation"):
                a = a.apply_transformation()
            else:
                pass
            args.append(a)
        return Shape(opc, *args)

    def complexity(self):
        """Gets complexity of shape.

        Returns
        -------
        complexity : int
            The complexity of the shape description. It is the number of
            surfaces needed to describe the shape. Repeats are taken into
            account.
        """
        args = self.args
        if len(args) == 1:
            return 1
        elif len(args) > 1:
            result = 0
            for a in args:
                result += a.complexity()
            return result
        else:
            return 0

    def get_surfaces(self) -> Set[Surface]:
        """Gets all the surfaces that describe the shape."""
        args = self.args
        if len(args) == 1:
            return {args[0]}
        elif len(args) > 1:
            result = set()
            for a in args:
                result = result.union(a.get_surfaces())
            return result
        else:
            return set()

    def is_empty(self):
        """Checks if the shape represents an empty set."""
        return self.opc == "E"

    def split_shape(self):
        shape_groups = []
        if self.opc == "U":
            stat = self.get_stat_table()
            drop_index = np.nonzero(np.all(stat == -1, axis=1))[0]
            arg_results = np.delete(stat, drop_index, axis=0)
            index_groups = self._find_groups(arg_results == +1)
            for ig in index_groups:
                index = np.nonzero(ig)[0]
                args = [self.args[i] for i in index]
                shape_groups.append(Shape("U", *args))
        elif self.opc == "I":
            arg_groups = [arg.split_shape() for arg in self.args]
            for args in product(*arg_groups):
                shape_groups.append(Shape("I", *args))
        else:
            shape_groups.append(self)
        return shape_groups

    @staticmethod
    def _find_groups(result):
        groups = [result[i, :] for i in range(result.shape[0])]
        while True:
            index = len(groups) - 1
            for j in range(index - 1, -1, -1):
                if np.any(groups[index] & groups[j] == 1):
                    groups[j] |= groups[index]
                    groups.pop(index)
                    break
            else:
                break
        return groups

    def get_simplest(self, trim_size=0):
        """Gets the simplest found description of the shape.

        Parameters
        ----------
        trim_size : int
            Shape variants with complexity greater than minimal one more than
            trim_size are thrown away.

        Returns
        -------
        shapes : list[Shape]
            A list of shapes with minimal complexity.
        """
        if self.opc != "I" and self.opc != "U":
            return [self]
        node_cases = []
        complexities = []
        stat = self.get_stat_table()
        if self.opc == "I":
            val = -1
        elif self.opc == "U":
            val = +1
        else:
            return {self}

        drop_index = np.nonzero(np.all(stat == -val, axis=1))[0]
        if len(drop_index) == 0:
            if self.opc == "I":
                return [Shape("E")]
            if self.opc == "U":
                return [Shape("R")]
        arg_results = np.delete(stat, drop_index, axis=0)
        if arg_results.shape[0] == 0:
            if self.opc == "I":
                return [Shape("R")]
            if self.opc == "U":
                return [Shape("E")]
        cases = self._find_coverages(arg_results, value=val)
        final_cases = set(tuple(c) for c in cases)
        if len(final_cases) == 0:
            print(self)
            return None
        unique = reduce(set.union, map(set, final_cases))
        args = self.args
        node_variants = {i: args[i].get_simplest(trim_size) for i in unique}
        for indices in final_cases:
            variants = [node_variants[i] for i in indices]
            for args in product(*variants):
                node = Shape(self.opc, *args)
                node_cases.append(node)
                complexities.append(node.complexity())
        sort_ind = np.argsort(complexities)
        final_nodes = []
        min_complexity = complexities[sort_ind[0]]
        for i in sort_ind:
            final_nodes.append(node_cases[i])
            if complexities[i] > min_complexity + trim_size:
                break
        return final_nodes

    @staticmethod
    def _find_coverages(results, value=+1):
        n = results.shape[1]
        cnt = np.count_nonzero(results == value, axis=1)
        i = np.argmin(cnt)
        cases = []
        for j in range(n):
            if results[i][j] == value:
                reminder = np.compress(results[:, j] != value, results, axis=0)
                if reminder.shape[0] == 0:
                    sub_cases = [[j]]
                else:
                    sub_cases = Shape._find_coverages(reminder, value=value)
                    for s in sub_cases:
                        s.append(j)
                cases.extend(sub_cases)
        for c in cases:
            c.sort()
        return cases

    def replace_surfaces(self, replace_dict):
        """Creates new Shape instance by replacing surfaces.

        If shape's surface is in replace dict, it is replaced by surface in
        dictionary values. Otherwise, the original surface is used. But new
        shape is created anyway.

        Parameters
        ----------
        replace_dict : dict
            A dictionary of surfaces to be replaced.

        Returns
        -------
        shape : Shape
            New Shape object obtained by replacing certain surfaces.
        """
        if self.opc == "C" or self.opc == "S":
            arg = self.args[0]
            surf = replace_dict.get(arg, arg)
            return Shape(self.opc, surf)
        elif self.opc == "I" or self.opc == "U":
            args = [arg.replace_surfaces(replace_dict) for arg in self.args]
            return Shape(self.opc, *args)
        else:
            return self

    @staticmethod
    def from_polish_notation(polish):
        """Creates Shape instance from reversed Polish notation.

        Parameters
        ----------
        polish : list
            List of surfaces and operations written in reversed Polish Notation.

        Returns
        -------
        shape : Shape
            The geometry represented by Shape instance.
        """
        operands = []
        for i, op in enumerate(polish):
            if isinstance(op, Surface):
                operands.append(Shape("S", op))
            elif isinstance(op, Shape):
                operands.append(op)
            elif op == "C":
                operands.append(operands.pop().complement())
            else:
                arg1 = operands.pop()
                arg2 = operands.pop()
                operands.append(Shape(op, arg1, arg2))
        return operands.pop()


def _clean_args(opc, *args):
    """Clean input arguments."""
    args = [a.shape if isinstance(a, Body) else a for a in args]
    _verify_opc(opc, *args)
    if opc == "I" or opc == "U":
        args = [Shape("S", a) if isinstance(a, Surface) else a for a in args]
    if len(args) > 1:
        # Extend arguments
        args = list(args)  # convert tuple to list
        i = 0
        while i < len(args):
            if args[i].opc == opc:
                a = args.pop(i)
                args.extend(a.args)
            else:
                i += 1

        i = 0
        while i < len(args):
            a = args[i]
            if a.opc == "E" and opc == "I" or a.opc == "R" and opc == "U":
                return a.opc, []
            elif a.opc == "E" and opc == "U" or a.opc == "R" and opc == "I":
                args.pop(i)
                continue
            for j, b in enumerate(args[i + 1 :]):
                if a.is_complement(b):
                    if opc == "I":
                        return "E", []
                    else:
                        return "R", []
            i += 1
        args = list(set(args))
        args.sort(key=hash)
        if len(args) == 0:
            opc = "E" if opc == "U" else "R"
    if len(args) == 1 and isinstance(args[0], Shape):
        if opc == "S" or opc == "I" or opc == "U":
            return args[0].opc, args[0].args
        if opc == "C":
            item = args[0].complement()
            return item.opc, item.args
    return opc, args


def _verify_opc(opc, *args):
    """Checks if such argument combination is valid."""
    if (opc == "E" or opc == "R") and len(args) > 0:
        raise ValueError("No arguments are expected.")
    elif (opc == "S" or opc == "C") and len(args) != 1:
        raise ValueError("Only one operand is expected.")
    elif opc == "I" or opc == "U":
        if len(args) == 0:
            raise ValueError("Operands are expected.")


TOperation = NewType("TOperation", str)
TGeometry = NewType("TGeometry", Union[List[Union[Surface, TOperation]], Shape, "Body"])

# noinspection PyProtectedMember


class Body(Card):
    """Represents MCNP cell.

    Parameters
    ----------
    geometry : list or Shape
        Geometry expression. It is either a list of Surface instances and
        operations (reverse Polish notation is used) or Shape object.
    options :
        A set of cell's options.

    Methods
    -------
    intersection(other)
        Returns an intersection of this cell with the other.
    fill(universe)
        Fills this cell by universe.
    simplify(box, split_disjoint, min_volume)
        Simplifies cell description.
    transform(tr)
        Applies transformation 'tr' to this cell.
    union(other)
        Returns a union of this cell with the other.
    """

    def __init__(self, geometry: TGeometry, **options) -> None:
        if isinstance(geometry, list):
            geometry = Shape.from_polish_notation(geometry)
        elif isinstance(geometry, Body):
            geometry = geometry.shape
        elif not isinstance(geometry, Shape):
            raise TypeError("Geometry list or Shape is expected.")
        Card.__init__(self, **options)
        self._shape = geometry

    def __repr__(self):
        options = str(self.options) if self.options else ""
        return f"Body({self._shape}, {options})"

    def __iter__(self):
        return iter(self._shape)

    def __hash__(self):
        return Card.__hash__(self) ^ hash(self._shape)

    def __eq__(self, other):
        return Card.__eq__(self, other) and self._shape == other._shape

    @property
    def transformation(self):
        return self.options.get("TRCL", None)

    def is_equivalent_to(self, other):
        result = self._shape == other._shape
        if result:
            if "FILL" in self.options:
                if "FILL" not in other.options:
                    return False
                my = self.options["FILL"]["universe"]
                their = other.options["FILL"]["universe"]
                return my.has_equivalent_cells(their)
        return result

    # TODO dvp: the method is used for printing, we'd better introduce virtual method print(self, out: TextIO)?
    # TODO dvp: in that case we could just return original text if available
    def mcnp_words(self, pretty=False):
        words = [str(self.name()), " "]
        if "MAT" in self.options.keys():
            words.append(str(self.options["MAT"].composition.name()))
            words.append(" ")
            words.append(str(-self.options["MAT"].density))
            words.append(" ")
        else:
            words.append("0")
            words.append(" ")
        words.extend(self._shape._get_words())
        words.append("\n")
        # insert options printing
        words.extend(self._options_list())
        for line in self.options.get("comment", []):
            words.append("$ " + str(line))
            words.append("\n")
        return words

    def _options_list(self):
        """Generates a list of option words.

        For __str__ method.
        """
        text = []
        for opt_group in CELL_OPTION_GROUPS:
            for key in opt_group:
                if key in self.options.keys():
                    text.extend(print_option(key, self.options[key]))
                    text.append(" ")
            text.append("\n")
        return text

    @property
    def shape(self):
        """Gets body's shape."""
        return self._shape

    def material(self) -> Optional[mm.Material]:
        """Gets body's Material.

        None is returned if no material present.
        """
        composition = self.options.get("MAT", None)
        assert composition is None or isinstance(composition, mm.Material)
        return composition

    def intersection(self, other):
        """Gets an intersection if this cell with the other.

        Other cell is a geometry that bounds this one. The resulting cell
        inherits all options of this one (the caller).

        Parameters
        ----------
        other : Cell
            Other cell.

        Returns
        -------
        cell : Cell
            The result.
        """
        geometry = self._shape.intersection(other)
        options = filter_dict(self.options, "original")
        return Body(geometry, **options)

    def union(self, other):
        """Gets a union if this cell with the other.

        The resulting cell inherits all options of this one (the caller).

        Parameters
        ----------
        other : Cell
            Other cell.

        Returns
        -------
        cell : Cell
            The result.
        """
        geometry = self._shape.union(other)
        options = filter_dict(self.options, "original")
        return Body(geometry, **options)

    def simplify(
        self,
        box=GLOBAL_BOX,
        split_disjoint=False,
        min_volume=MIN_BOX_VOLUME,
        trim_size=1,
    ):
        """Simplifies this cell by removing unnecessary surfaces.

        The simplification procedure goes in the following way.
        # TODO: insert brief description!

        Parameters
        ----------
        box : Box
            Box where geometry should be simplified.
        split_disjoint : bool
            Whether to split disjoint geometries into separate geometries.
        min_volume : float
            The smallest value of box's volume when the process of box splitting
            must be stopped.
        trim_size : int
            Max size of set to return. It is used to prevent unlimited growth
            of the variant set.

        Returns
        -------
        simple_cell : Cell
            Simplified version of this cell.
        """
        # print('Collect stage...')
        self._shape.collect_statistics(box, min_volume)
        # print('finding optimal solution...')
        variants = self._shape.get_simplest(trim_size)
        # print(len(variants))
        options = filter_dict(self.options, "original")
        return Body(variants[0], **options)

    def split(self, box=GLOBAL_BOX, min_volume=MIN_BOX_VOLUME):
        """Splits cell into disjoint cells.

        Returns
        -------
        cells : list
        """
        self.shape.collect_statistics(box, min_volume)
        shape_groups = self.shape.split_shape()
        bodies = [Body(shape, **self.options) for shape in shape_groups]
        return bodies

    # noinspection PyShadowingNames
    def fill(
        self,
        universe=None,
        recurrent: bool = False,
        simplify: bool = False,
        **kwargs: dict[str, any],
    ) -> list["Body"]:
        """Fills this cell by filling universe.

        If this cell doesn't contain fill options and universe is not
        specified, the cell itself is returned as list of length 1. Otherwise:
        a list of cells from filling universe bounded by cell being filled is
        returned.

        Args:
            universe:
                Universe which cells fill this one. If None, universe from 'FILL'
                option will be used. If no such universe, the cell itself will be
                returned. Default: None.
            recurrent:
                If filler universe also contains cells with fill option, they will
                be also filled. Default: False.
            simplify:
                If True, all cells obtained will be simplified.
            **kwargs: dict
                Keyword parameters for simplify method if simplify is True.
                Default: all False.

        Returns:
            The list of resulting cells.
        """
        if universe is None:
            if "FILL" in self.options.keys():
                universe = self.options["FILL"]["universe"]
                tr = self.options["FILL"].get("transform", None)
                if tr:
                    universe = universe.transform(tr)
            else:
                return [self]
        if recurrent:
            universe = universe.fill(recurrent=True, simplify=simplify, **kwargs)
        cells = []
        for c in universe:
            new_cell = c.intersection(self)  # because properties like MAT, etc
            # must be as in filling cell.
            if "U" in self.options.keys():
                new_cell.options["U"] = self.options["U"]  # except universe.
            if simplify:
                new_cell = new_cell.simplify(**kwargs)
            cells.append(new_cell)
        return cells

    def transform(self, tr):
        """Applies transformation to this cell.

        Parameters
        ----------
        tr : Transform
            Transformation to be applied.

        Returns
        -------
        cell : Cell
            The result of this cell transformation.
        """
        geometry = self._shape.transform(tr)
        options = filter_dict(self.options, "original")
        cell = Body(geometry, **options)
        fill = cell.options.get("FILL", None)
        if fill:
            tr_in = fill.get("transform", Transformation())
            new_tr = tr.apply2transform(tr_in)
            fill["transform"] = new_tr
        return cell

    def apply_transformation(self) -> "Body":
        """Actually apply transformation to this cell."""
        geometry = self._shape.apply_transformation()
        options = filter_dict(self.options, "original")
        cell = Body(geometry, **options)
        fill = cell.options.get("FILL", None)
        if fill:
            tr_in = fill.get("transform", None)
            filling_universe = fill["universe"]
            new_filling_universe = (
                deepcopy(filling_universe).transform(tr_in).apply_transformation()
            )
            cell.options["FILL"] = {"universe": new_filling_universe}
            # TODO dvp: this should create a lot of cell clashes on complex models with multiple filling with
            #           one universe. Should be resolved before saving.
        return cell


def simplify(cells: Iterable, box: Box = GLOBAL_BOX, min_volume: float = 1.0) -> tp.Generator:
    """Simplifies the cells.

    Parameters
    ----------

    cells:
        iterable over cells to simplify
    box :
        Box, from which simplification process starts. Default: GLOBAL_BOX.
    min_volume : float
        Minimal volume of the box, when splitting process terminates.
    """

    for c in cells:
        cs = c.simplify(box=box, min_volume=min_volume)
        if not cs.shape.is_empty():
            yield cs


class Simplifier(object):
    def __init__(self, box: Box = GLOBAL_BOX, min_volume: float = 1.0):
        self.box = box
        self.min_volume = min_volume

    def __call__(self, cell: Body):
        return cell.simplify(box=self.box, min_volume=self.min_volume)

    def __getstate__(self):
        return self.box, self.min_volume

    def __setstate__(self, state):
        box, min_volume = state
        self.__init__(box, min_volume)


def simplify_mp(
    cells: Iterable[Body], box: Box = GLOBAL_BOX, min_volume: float = 1.0, chunk_size=1
) -> tp.Generator:
    """Simplifies the cells in multiprocessing mode.

    Parameters
    ----------

    cells:
        iterable over cells to simplify
    box :
        Box, from which simplification process starts. Default: GLOBAL_BOX.
    min_volume : float
        Minimal volume of the box, when splitting process terminates.
    chunk_size: size of chunks to pass to child processes
    """
    cpus = os.cpu_count()
    with Pool(processes=cpus) as pool:
        yield from pool.imap(
            Simplifier(box=box, min_volume=min_volume), cells, chunksize=chunk_size
        )


def simplify_mpp(
    cells: Iterable[Body],
    box: Box = GLOBAL_BOX,
    min_volume: float = 1.0,
    chunk_size: int = 1,
) -> tp.Generator:
    """Simplifies the cells in multiprocessing mode with progress bar.

    Parameters
    ----------

    cells:
        iterable over cells to simplify
    box :
        Box, from which simplification process starts. Default: GLOBAL_BOX.
    min_volume : float
        Minimal volume of the box, when splitting process terminates.
    chunk_size: size of chunks to pass to child processes
    """

    def fmt_fun(x):
        return "Simplifying cell #{0}".format(x.name() if x else x)

    with progressbar(simplify_mp(cells, box, min_volume, chunk_size), item_show_func=fmt_fun) as pb:
        for c in pb:
            yield c



"""Provide package version and other meta information."""
try:
    import importlib_metadata as meta
except ImportError:
    import importlib.metadata as meta  # type: ignore

__title__ = "mckit"
__distribution__ = meta.distribution(__title__)
__meta_data__ = __distribution__.metadata
__author__ = __meta_data__["Author"]
__author_email__ = __meta_data__["Author-email"]
__license__ = __meta_data__["License"]
__summary__ = __meta_data__["Summary"]
__copyright__ = "Copyright 2018-2020 Roman Rodionov"  # TODO dvp: move to meta (project.toml)
__version__ = __distribution__.version



# -*- coding: utf-8 -*-
from pathlib import Path

from mckit.cli.runner import mckit
from mckit.utils.resource import filename_resolver

data_filename_resolver = filename_resolver("tests.cli")


def test_when_there_is_no_args(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["concat"], catch_exceptions=False)
        assert result.exit_code != 0, "Should fail when no arguments provided"
        assert "Usage:" in result.output


def test_not_existing_file(runner):
    result = runner.invoke(mckit, args=["concat", "not-existing.txt"], catch_exceptions=False)
    assert result.exit_code > 0
    assert "Path 'not-existing.txt' does not exist" in result.output


def test_when_only_part_is_specified(runner):
    part = data_filename_resolver("data/concat/test_load_table_1.csv")
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["concat", part], catch_exceptions=False)
        assert result.exit_code == 0, "Should success without specified output: " + result.output
        assert (
            "x   y" in result.output
        ), "Should send output to stdout, when the output is not specified"


def test_when_output_is_specified(runner):
    part = data_filename_resolver("data/concat/test_load_table_1.csv")
    with runner.isolated_filesystem() as prefix:
        output_file = Path(prefix) / "test_when_output_is_specified.txt"
        result = runner.invoke(
            mckit,
            args=["concat", "--output", str(output_file), part],
            catch_exceptions=False,
        )
        assert result.exit_code == 0, "Should success with specified output: " + result.output
        assert output_file.exists(), "Should create output file " + output_file
        # noinspection PyCompatibility
        assert "x   y" in output_file.read_text(
            encoding="Cp1251"
        ), f"Should contain content of '{part}'"


# noinspection PyCompatibility
def test_when_two_parts_are_specified(runner):
    part1 = data_filename_resolver("data/concat/test_load_table_1.csv")
    part2 = data_filename_resolver("data/concat/test_load_table_2.csv")
    with runner.isolated_filesystem() as prefix:
        output_file = Path(prefix) / "test_when_output_is_specified.txt"
        result = runner.invoke(
            mckit,
            args=["concat", "--output", str(output_file), part1, part2],
            catch_exceptions=False,
        )
        assert result.exit_code == 0, "Should success with specified output: " + result.output
        assert output_file.exists(), "Should create output file " + output_file
        text = output_file.read_text(encoding="Cp1251")
        assert "x   y" in text, f"Should contain content of '{part1}'"
        assert "x    ;   y" in text, f"Should contain content of '{part2}'"


def test_when_output_file_exists_and_override_is_not_specified(runner):
    part = data_filename_resolver("data/concat/test_load_table_1.csv")
    with runner.isolated_filesystem() as prefix:
        output_file = Path(prefix) / "test_when_output_is_specified.txt"
        output_file.touch(exist_ok=False)
        result = runner.invoke(
            mckit, args=["concat", "-o", str(output_file), part], catch_exceptions=False
        )
        assert (
            result.exit_code != 0
        ), "Should fail when output file exist and override is not specified"



import numpy as np
import pytest

from mckit.geometry import ORIGIN
from mckit.transformation import IDENTITY_ROTATION, Transformation


@pytest.mark.parametrize(
    "args, rot, offset, options",
    [
        ({}, IDENTITY_ROTATION, ORIGIN, {}),
        ({"indegrees": True}, IDENTITY_ROTATION, ORIGIN, {}),
        ({"indegrees": False}, IDENTITY_ROTATION, ORIGIN, {}),
        ({"inverted": True}, IDENTITY_ROTATION, ORIGIN, {}),
        ({"inverted": False}, IDENTITY_ROTATION, ORIGIN, {}),
        ({"translation": [1, 0, 0]}, IDENTITY_ROTATION, np.array([1, 0, 0]), {}),
        (
            {"translation": [1, 0, 0], "inverted": True},
            IDENTITY_ROTATION,
            np.array([-1, 0, 0]),
            {},
        ),
        (
            {"translation": [1, 2, 3], "inverted": True},
            IDENTITY_ROTATION,
            np.array([-1, -2, -3]),
            {},
        ),
        (
            {"translation": [1, 2, 3], "inverted": True, "name": 1},
            IDENTITY_ROTATION,
            np.array([-1, -2, -3]),
            {"name": 1},
        ),
        (
            {"rotation": np.cos(np.array([30, 60, 90, 120, 30, 90, 90, 90, 0]) * np.pi / 180)},
            np.cos(np.array([[30, 120, 90], [60, 30, 90], [90, 90, 0]]) * np.pi / 180),
            ORIGIN,
            {},
        ),
        (
            {"rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0], "indegrees": True},
            np.cos(np.array([[30, 120, 90], [60, 30, 90], [90, 90, 0]]) * np.pi / 180),
            ORIGIN,
            {},
        ),
        (
            {
                "rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0],
                "indegrees": True,
                "translation": [1, 2, 3],
            },
            np.cos(np.array([[30, 120, 90], [60, 30, 90], [90, 90, 0]]) * np.pi / 180),
            np.array([1, 2, 3]),
            {},
        ),
        (
            {
                "rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0],
                "indegrees": True,
                "translation": [1, 2, 3],
                "inverted": True,
            },
            np.cos(np.array([[30, 120, 90], [60, 30, 90], [90, 90, 0]]) * np.pi / 180),
            np.array([-(np.sqrt(3) - 2) / 2, -(2 * np.sqrt(3) + 1) / 2, -3]),
            {},
        ),
    ],
)
def test_creation(args, rot, offset, options):
    tr = Transformation(**args)
    np.testing.assert_array_almost_equal(tr._u, rot)
    np.testing.assert_array_almost_equal(tr._t, offset)
    for k, v in options.items():
        assert tr[k] == v


@pytest.mark.parametrize(
    "args",
    [
        {"translation": [0]},
        {"translation": [0, 1]},
        {"translation": [0, 1, 2, 3]},
        {"translation": [0, 1, 2, 3, 4]},
        {"rotation": [0, 1, 2]},
        {"rotation": [1, 2, 3, 4, 5, 6, 7, 8]},
        {"rotation": [30.058, 59.942, 90, 120, 30, 90, 90, 90, 0], "indegrees": True},
        {"rotation": [29.942, 60.058, 90, 120, 30, 90, 90, 90, 0], "indegrees": True},
        {"rotation": [30, 60, 90, 120.058, 30.058, 90, 90, 90, 0], "indegrees": True},
        {"rotation": [30, 60, 90, 119.942, 29.942, 90, 90, 90, 0], "indegrees": True},
        {
            "rotation": [30, 60, 90, 120, 30, 90, 90.058, 90.058, 0.058],
            "indegrees": True,
        },
        {
            "rotation": [30, 60, 90, 120, 30, 90, 89.942, 89.942, 0.058],
            "indegrees": True,
        },
    ],
)
def test_creation_failure(args):
    with pytest.raises(ValueError):
        Transformation(**args)


@pytest.mark.parametrize(
    "args, answer",
    [
        (
            {"rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0], "indegrees": True},
            [0, 0, 0, 30, 60, 90, 120, 30, 90, 90, 90, 0],
        ),
        (
            {
                "rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0],
                "indegrees": True,
                "translation": [1, 2, 3],
            },
            [1, 2, 3, 30, 60, 90, 120, 30, 90, 90, 90, 0],
        ),
    ],
)
def test_words(args, answer):
    tr = Transformation(**args)
    words = [float(w) for w in "".join(tr.get_words()).split()]
    np.testing.assert_array_almost_equal(words, answer)


@pytest.mark.parametrize(
    "args, options, answer",
    [
        (
            {"rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0], "indegrees": True},
            {},
            None,
        ),
        (
            {
                "rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0],
                "indegrees": True,
                "translation": [1, 2, 3],
            },
            {"name": 1},
            1,
        ),
        (
            {
                "rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0],
                "indegrees": True,
                "translation": [1, 2, 3],
            },
            {"name": 2},
            2,
        ),
        (
            {
                "rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0],
                "indegrees": True,
                "translation": [1, 2, 3],
            },
            {"name": 3},
            3,
        ),
    ],
)
def test_name(args, options, answer):
    tr = Transformation(**args, **options)
    assert tr.name() == answer


@pytest.mark.parametrize(
    "args",
    [
        {"rotation": [30.057, 59.943, 90, 120, 30, 90, 90, 90, 0], "indegrees": True},
        {"rotation": [29.943, 60.057, 90, 120, 30, 90, 90, 90, 0], "indegrees": True},
        {"rotation": [30, 60, 90, 120.057, 30.057, 90, 90, 90, 0], "indegrees": True},
        {"rotation": [30, 60, 90, 119.943, 29.943, 90, 90, 90, 0], "indegrees": True},
        {"rotation": [30, 60, 90, 120, 30, 90, 90.057, 90, 0.057], "indegrees": True},
        {"rotation": [30, 60, 90, 120, 30, 90, 89.943, 90, 0.057], "indegrees": True},
    ],
)
def test_orthogonalization(args):
    tr = Transformation(**args)
    result = np.dot(tr._u.transpose(), tr._u)
    np.testing.assert_array_almost_equal(result, IDENTITY_ROTATION)


@pytest.fixture(scope="module")
def transforms():
    return Transformation(
        translation=[1, 2, -3],
        indegrees=True,
        rotation=[30, 60, 90, 120, 30, 90, 90, 90, 0],
    )


@pytest.mark.parametrize(
    "points, expected",
    [
        (np.array([1, 0, 0]), np.array([np.sqrt(3) / 2 + 1, 2.5, -3])),
        (np.array([2, 0, 0]), np.array([np.sqrt(3) + 1, 3, -3])),
        (np.array([0, 1, 0]), np.array([0.5, np.sqrt(3) / 2 + 2, -3])),
        (np.array([0, 2, 0]), np.array([0, np.sqrt(3) + 2, -3])),
        (np.array([0, 0, 1]), np.array([1, 2, -2])),
        (
            np.array([[1, 0, 0], [2, 0, 0], [0, 1, 0], [0, 2, 0], [0, 0, 1]]),
            np.array(
                [
                    [np.sqrt(3) / 2 + 1, 2.5, -3],
                    [np.sqrt(3) + 1, 3, -3],
                    [0.5, np.sqrt(3) / 2 + 2, -3],
                    [0, np.sqrt(3) + 2, -3],
                    [1, 2, -2],
                ]
            ),
        ),
    ],
)
def test_point_transformation(transforms, points, expected):
    result = transforms.apply2point(points)
    np.testing.assert_array_almost_equal(result, expected)


@pytest.mark.parametrize(
    "vectors, expected",
    [
        (np.array([1, 0, 0]), np.array([np.sqrt(3) / 2, 0.5, 0])),
        (np.array([2, 0, 0]), np.array([np.sqrt(3), 1, 0])),
        (np.array([0, 1, 0]), np.array([-0.5, np.sqrt(3) / 2, 0])),
        (np.array([0, 2, 0]), np.array([-1, np.sqrt(3), 0])),
        (np.array([0, 0, 1]), np.array([0, 0, 1])),
        (
            np.array([[1, 0, 0], [2, 0, 0], [0, 1, 0], [0, 2, 0], [0, 0, 1]]),
            np.array(
                [
                    [np.sqrt(3) / 2, 0.5, 0],
                    [np.sqrt(3), 1, 0],
                    [-0.5, np.sqrt(3) / 2, 0],
                    [-1, np.sqrt(3), 0],
                    [0, 0, 1],
                ]
            ),
        ),
    ],
)
def test_vector_transformation(transforms, vectors, expected):
    result = transforms.apply2vector(vectors)
    np.testing.assert_array_almost_equal(result, expected)


@pytest.mark.parametrize(
    "norm, point",
    [
        (np.array([1, 0, 0]), np.array([2, 3, 5])),
        (np.array([0, 1, 0]), np.array([2, 3, 5])),
        (np.array([0, 0, 1]), np.array([2, 3, 5])),
        (np.array([1, -2, 3]), np.array([2, 3, 5])),
        (np.array([-4, 0, 5]), np.array([2, 3, 5])),
    ],
)
def test_plane_transformation(transforms, norm, point):
    k1 = -np.dot(norm, point)
    v, k = transforms.apply2plane(norm, k1)
    v_ref = transforms.apply2vector(norm)
    k_ref = -np.dot(v_ref, transforms.apply2point(point))
    np.testing.assert_array_almost_equal(v, v_ref)
    assert k == pytest.approx(k_ref, rel=1.0e-10)


@pytest.mark.parametrize("point, radius", [(np.array([3, -4, 7]), 4)])
def test_gq_transformation(transforms, point, radius):
    m1 = IDENTITY_ROTATION
    v1 = -2 * point
    k1 = np.linalg.norm(point) ** 2 - radius**2
    m, v, k = transforms.apply2gq(m1, v1, k1)
    v_ref = -2 * transforms.apply2point(point)
    k_ref = np.linalg.norm(-0.5 * v_ref) ** 2 - radius**2
    np.testing.assert_array_almost_equal(m, m1)
    np.testing.assert_array_almost_equal(v, v_ref)
    np.testing.assert_array_almost_equal(k, k_ref)


tr_tr_cases = [
    {"translation": [1, 2, 3]},
    {"translation": [-4, 2, -3]},
    {
        "translation": [3, 0, 9],
        "rotation": [30, 60, 90, 120, 30, 90, 90, 90, 0],
        "indegrees": True,
    },
    {
        "translation": [1, 4, -2],
        "rotation": [0, 90, 90, 90, 30, 60, 90, 120, 30],
        "indegrees": True,
    },
    {
        "translation": [-2, 5, 3],
        "rotation": [30, 90, 60, 90, 0, 90, 120, 90, 30],
        "indegrees": True,
    },
]


@pytest.fixture(scope="function")
def trtr():
    tr = [Transformation(**tdata) for tdata in tr_tr_cases]
    return tr


@pytest.mark.parametrize("tr1_no", range(len(tr_tr_cases)))
@pytest.mark.parametrize("tr2_no", range(len(tr_tr_cases)))
@pytest.mark.parametrize(
    "point",
    [
        np.array([1, 0, 0]),
        np.array([2, -3, 1]),
        np.array([-4, 1, 9]),
        np.array([7, 2, 5]),
        np.array([8, -1, 3]),
        np.array([8, -3, 2]),
        np.array([3, 6, 4]),
        np.array([2, -5, -1]),
        np.array([-2, 7, 2]),
    ],
)
def test_apply2tr(trtr, tr1_no, tr2_no, point):
    tr1 = trtr[tr1_no]
    tr2 = trtr[tr2_no]
    ans = tr1.apply2point(tr2.apply2point(point))
    tr = tr1.apply2transform(tr2)
    result = tr.apply2point(point)
    np.testing.assert_array_almost_equal(ans, result)


@pytest.mark.parametrize(
    "p1, p",
    [
        (np.array([1, 0, 0]), np.array([np.sqrt(3) / 2 + 1, 2.5, -3])),
        (np.array([2, 0, 0]), np.array([np.sqrt(3) + 1, 3, -3])),
        (np.array([0, 1, 0]), np.array([0.5, np.sqrt(3) / 2 + 2, -3])),
        (np.array([0, 2, 0]), np.array([0, np.sqrt(3) + 2, -3])),
        (np.array([0, 0, 1]), np.array([1, 2, -2])),
        (
            np.array([[1, 0, 0], [2, 0, 0], [0, 1, 0], [0, 2, 0], [0, 0, 1]]),
            np.array(
                [
                    [np.sqrt(3) / 2 + 1, 2.5, -3],
                    [np.sqrt(3) + 1, 3, -3],
                    [0.5, np.sqrt(3) / 2 + 2, -3],
                    [0, np.sqrt(3) + 2, -3],
                    [1, 2, -2],
                ]
            ),
        ),
    ],
)
def test_reverse(transforms, p1, p):
    tr_inv = transforms.reverse()
    result = tr_inv.apply2point(p)
    np.testing.assert_array_almost_equal(result, p1)


@pytest.mark.parametrize(
    "options", [{"name": 1, "comment": "abcdef"}, {"name": 1}, {"comment": "rrreee"}]
)
@pytest.mark.parametrize("tr_no", range(len(tr_tr_cases)))
def test_set_item(trtr, tr_no, options):
    tr = trtr[tr_no]
    for k, v in options.items():
        tr[k] = v
    for k, v in options.items():
        assert tr[k] == v
